# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, cast

from keystoneauth1 import adapter

from openstack import resource
from openstack import utils


class Instance(resource.Resource):
    resource_key = 'instance'
    resources_key = 'instances'
    base_path = '/instances'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    # Properties
    #: The flavor of the instance
    flavor = resource.Body('flavor')
    #: Links associated with the instance
    links = resource.Body('links')
    #: The name of the instance
    name = resource.Body('name')
    #: The status of the instance
    status = resource.Body('status')
    #: The size of the volume
    volume = resource.Body('volume')
    #: A dictionary of datastore details, often including 'type' and 'version'
    #: keys
    datastore = resource.Body('datastore', type=dict)
    #: The ID of this instance
    id = resource.Body('id')
    #: The region this instance resides in
    region = resource.Body('region')
    #: The name of the host
    hostname = resource.Body('hostname')
    #: The timestamp when this instance was created
    created_at = resource.Body('created')
    #: The timestamp when this instance was updated
    updated_at = resource.Body('updated')

    def enable_root_user(self, session: adapter.Adapter) -> dict[str, Any]:
        """Enable login for the root user.

        This operation enables login from any host for the root user
        and provides the user with a generated root password.

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :returns: A dictionary with keys ``name`` and ``password`` specifying
            the login credentials.
        """
        url = utils.urljoin(self.base_path, self.id, 'root')
        resp = session.post(url)
        return cast(dict[str, Any], resp.json()['user'])

    def is_root_enabled(self, session: adapter.Adapter) -> bool:
        """Determine if root is enabled on an instance.

        Determine if root is enabled on this particular instance.

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :returns: ``True`` if root user is enabled for a specified database
            instance or ``False`` otherwise.
        """
        url = utils.urljoin(self.base_path, self.id, 'root')
        resp = session.get(url)
        return cast(bool, resp.json()['rootEnabled'])

    def restart(self, session: adapter.Adapter) -> None:
        """Restart the database instance

        :returns: ``None``
        """
        body = {'restart': None}
        url = utils.urljoin(self.base_path, self.id, 'action')
        session.post(url, json=body)

    def resize(self, session: adapter.Adapter, flavor_reference: str) -> None:
        """Resize the database instance

        :returns: ``None``
        """
        body = {'resize': {'flavorRef': flavor_reference}}
        url = utils.urljoin(self.base_path, self.id, 'action')
        session.post(url, json=body)

    def resize_volume(
        self, session: adapter.Adapter, volume_size: int
    ) -> None:
        """Resize the volume attached to the instance

        :returns: ``None``
        """
        body = {'resize': {'volume': volume_size}}
        url = utils.urljoin(self.base_path, self.id, 'action')
        session.post(url, json=body)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Iterator
import copy
import importlib.metadata as importlib_metadata
import os.path
from typing import (
    Any,
    Optional,
    Protocol,
    TYPE_CHECKING,
    TypeVar,
    cast,
    overload,
)
from urllib import parse
import warnings

try:
    import keyring
except ImportError:
    keyring = None  # type: ignore[assignment]

from keystoneauth1.access import service_catalog as ks_service_catalog
from keystoneauth1 import discover
import keystoneauth1.exceptions.catalog
from keystoneauth1.identity import base as ks_identity_base
from keystoneauth1.loading import adapter as ks_load_adap
from keystoneauth1 import plugin
from keystoneauth1 import session as ks_session
import os_service_types
import urllib3.exceptions

try:
    import statsd as statsd_client
except ImportError:
    statsd_client = None
try:
    import prometheus_client
except ImportError:
    prometheus_client = None  # type: ignore[assignment]
try:
    # NOTE(stephenfin): This library is EOL so we explicitly don't have it in
    # our dependencies
    import influxdb as influxdb_client  # type: ignore[import-not-found]
except ImportError:
    influxdb_client = None

from openstack import _log
from openstack.config import _util
from openstack.config import defaults as config_defaults
from openstack import exceptions
from openstack import proxy
from openstack import warnings as os_warnings

if TYPE_CHECKING:
    from oslo_config import cfg

    from openstack.config import loader


_T = TypeVar('_T')

_logger = _log.setup_logging('openstack')

SCOPE_KEYS = {
    'domain_id',
    'domain_name',
    'project_id',
    'project_name',
    'system_scope',
}

# Sentinel for nonexistence
_ENOENT = object()


class _PasswordCallback(Protocol):
    def __call__(self, prompt: str | None = None) -> str: ...


def _make_key(key: str, service_type: str | None) -> str:
    if not service_type:
        return key
    else:
        service_type = service_type.lower().replace('-', '_')
        return "_".join([service_type, key])


def _disable_service(
    config: dict[str, Any],
    service_type: str,
    reason: str | None = None,
) -> None:
    service_type = service_type.lower().replace('-', '_')
    key = f'has_{service_type}'
    config[key] = False
    if reason:
        d_key = _make_key('disabled_reason', service_type)
        config[d_key] = reason


def _get_implied_microversion(version: str | None) -> str | None:
    if version and '.' in version:
        # Some services historically had a .0 in their normal api version.
        # Neutron springs to mind with version "2.0". If a user has "2.0"
        # set in a variable or config file just because history, we don't
        # need to send any microversion headers.
        if version.split('.')[1] != "0":
            return version

    return None


def from_session(
    session: ks_session.Session,
    name: str | None = None,
    region_name: str | None = None,
    force_ipv4: bool = False,
    app_name: str | None = None,
    app_version: str | None = None,
    **kwargs: Any,
) -> 'CloudRegion':
    """Construct a CloudRegion from an existing `keystoneauth1.session.Session`

    When a Session already exists, we don't actually even need to go through
    the OpenStackConfig.get_one_cloud dance. We have a Session with Auth info.
    The only parameters that are really needed are adapter/catalog related.

    :param keystoneauth1.session.session session:
        An existing authenticated Session to use.
    :param str name:
        A name to use for this cloud region in logging. If left empty, the
        hostname of the auth_url found in the Session will be used.
    :param str region_name:
        The region name to connect to.
    :param bool force_ipv4:
        Whether or not to disable IPv6 support. Defaults to False.
    :param str app_name:
        Name of the application to be added to User Agent.
    :param str app_version:
        Version of the application to be added to User Agent.
    :param kwargs:
        Config settings for this cloud region.
    """
    config_dict = config_defaults.get_defaults()
    config_dict.update(**kwargs)
    return CloudRegion(
        name=name,
        session=session,
        config=config_dict,
        region_name=region_name,
        force_ipv4=force_ipv4,
        app_name=app_name,
        app_version=app_version,
    )


def from_conf(
    conf: 'cfg.ConfigOpts',
    session: ks_session.Session | None = None,
    service_types: list[str] | None = None,
    **kwargs: Any,
) -> 'CloudRegion':
    """Create a CloudRegion from oslo.config ConfigOpts.

    :param oslo_config.cfg.ConfigOpts conf:
        An oslo.config ConfigOpts containing keystoneauth1.Adapter options in
        sections named according to project (e.g. [nova], not [compute]).
        TODO: Current behavior is to use defaults if no such section exists,
        which may not be what we want long term.
    :param keystoneauth1.session.Session session:
        An existing authenticated Session to use. This is currently required.
        TODO: Load this (and auth) from the conf.
    :param service_types:
        A list/set of service types for which to look for and process config
        opts. If None, all known service types are processed. Note that we will
        not error if a supplied service type can not be processed successfully
        (unless you try to use the proxy, of course). This tolerates uses where
        the consuming code has paths for a given service, but those paths are
        not exercised for given end user setups, and we do not want to generate
        errors for e.g. missing/invalid conf sections in those cases. We also
        don't check to make sure your service types are spelled correctly -
        caveat implementor.
    :param kwargs:
        Additional keyword arguments to be passed directly to the CloudRegion
        constructor.
    :raise openstack.exceptions.ConfigException:
        If session is not specified.
    :return:
        An openstack.config.cloud_region.CloudRegion.
    """
    if not session:
        # TODO(mordred) Fill this in - not needed for first stab with nova
        raise exceptions.ConfigException("A Session must be supplied.")
    config_dict = kwargs.pop('config', config_defaults.get_defaults())
    stm = os_service_types.ServiceTypes()
    for st in stm.all_types_by_service_type:
        if service_types is not None and st not in service_types:
            _disable_service(
                config_dict,
                st,
                reason="Not in the list of requested service_types.",
            )
            continue
        project_name = stm.get_project_name(st)
        if project_name is None:
            _disable_service(
                config_dict,
                st,
                reason=f"No project name found for service type '{st}'.",
            )
            continue
        if project_name not in conf:
            if '-' in project_name:
                project_name = project_name.replace('-', '_')

            if project_name not in conf:
                _disable_service(
                    config_dict,
                    st,
                    reason=(
                        f"No section for project '{project_name}' (service "
                        f"type '{st}') was present in the config."
                    ),
                )
                continue
        opt_dict: dict[str, str] = {}
        # Populate opt_dict with (appropriately processed) Adapter conf opts
        try:
            ks_load_adap.process_conf_options(conf[project_name], opt_dict)
        except Exception as e:
            # NOTE(efried): This is for (at least) a couple of scenarios:
            # (1) oslo_config.cfg.NoSuchOptError when ksa adapter opts are not
            #     registered in this section.
            # (2) TypeError, when opts are registered but bogus (e.g.
            #     'interface' and 'valid_interfaces' are both present).
            # We may want to consider (providing a kwarg giving the caller the
            # option of) blowing up right away for (2) rather than letting them
            # get all the way to the point of trying the service and having
            # *that* blow up.
            reason = (
                "Encountered an exception attempting to process config "
                f"for project '{project_name}' (service type "
                f"'{st}'): {e}"
            )
            _logger.warning(f"Disabling service '{st}': {reason}")
            _disable_service(config_dict, st, reason=reason)
            continue
        # Load them into config_dict under keys prefixed by ${service_type}_
        for raw_name, opt_val in opt_dict.items():
            config_name = _make_key(raw_name, st)
            config_dict[config_name] = opt_val
    return CloudRegion(session=session, config=config_dict, **kwargs)


class CloudRegion:
    # TODO(efried): Doc the rest of the kwargs
    """The configuration for a Region of an OpenStack Cloud.

    A CloudRegion encapsulates the config information needed for connections
    to all of the services in a Region of a Cloud.

    :param name:
    :param region_name:
        The default region name for all services in this CloudRegion. If
        both ``region_name`` and ``config['region_name']`` are specified, the
        kwarg takes precedence. May be overridden for a given ${service}
        via a ${service}_region_name key in the ``config`` dict.
    :param config:
        A dict of configuration values for the CloudRegion and its
        services. The key for a ${config_option} for a specific ${service}
        should be ${service}_${config_option}. For example, to configure
        the endpoint_override for the block_storage service, the ``config``
        dict should contain::

            'block_storage_endpoint_override': 'http://...'

        To provide a default to be used if no service-specific override is
        present, just use the unprefixed ${config_option} as the service
        key, e.g.::

            'interface': 'public'
    :param force_ipv4:
    :param auth_plugin:
    :param openstack_config:
    :param session_constructor:
    :param app_name:
    :param app_version:
    :param session:
    :param discovery_cache:
    :param extra_config:
    :param cache_expiration_time:
    :param cache_expirations:
    :param cache_path:
    :param cache_class:
    :param cache_arguments:
    :param password_callback:
    :param statsd_host:
    :param statsd_port:
    :param statsd_prefix:
    :param influxdb_config:
    :param collector_registry:
    :param cache_auth:
    """

    def __init__(
        self,
        name: str | None = None,
        region_name: str | None = None,
        config: dict[str, Any] | None = None,
        force_ipv4: bool = False,
        auth_plugin: plugin.BaseAuthPlugin | None = None,
        openstack_config: Optional['loader.OpenStackConfig'] = None,
        session_constructor: type[ks_session.Session] | None = None,
        app_name: str | None = None,
        app_version: str | None = None,
        session: ks_session.Session | None = None,
        discovery_cache: dict[str, discover.Discover] | None = None,
        extra_config: dict[str, Any] | None = None,
        cache_expiration_time: int = 0,
        cache_expirations: dict[str, int] | None = None,
        cache_path: str | None = None,
        cache_class: str = 'dogpile.cache.null',
        cache_arguments: dict[str, Any] | None = None,
        password_callback: _PasswordCallback | None = None,
        statsd_host: str | None = None,
        statsd_port: str | None = None,
        statsd_prefix: str | None = None,
        # TODO(stephenfin): Add better types
        influxdb_config: dict[str, Any] | None = None,
        collector_registry: Optional[
            'prometheus_client.CollectorRegistry'
        ] = None,
        cache_auth: bool = False,
    ) -> None:
        self._name = name
        self.config = _util.normalize_keys(config or {})
        # NOTE(efried): For backward compatibility: a) continue to accept the
        # region_name kwarg; b) make it take precedence over (non-service_type-
        # specific) region_name set in the config dict.
        if region_name is not None:
            self.config['region_name'] = region_name
        self._extra_config = extra_config or {}
        self.log = _log.setup_logging('openstack.config')
        self._force_ipv4 = force_ipv4
        self._auth = auth_plugin
        self._cache_auth = cache_auth
        self.load_auth_from_cache()
        self._openstack_config = openstack_config
        self._keystone_session = session
        self._session_constructor = session_constructor or ks_session.Session
        self._app_name = app_name
        self._app_version = app_version
        self._discovery_cache = discovery_cache or None
        self._cache_expiration_time = cache_expiration_time
        self._cache_expirations = cache_expirations or {}
        self._cache_path = cache_path
        self._cache_class = cache_class
        self._cache_arguments = cache_arguments
        self._password_callback = password_callback
        self._statsd_host = statsd_host
        self._statsd_port = statsd_port
        self._statsd_prefix = statsd_prefix
        self._statsd_client = None
        self._influxdb_config = influxdb_config
        self._influxdb_client = None

        if influxdb_config is not None:
            # NOTE(stephenfin): If you are a user and care about InfluxDB,
            # please propose patches to migrate this to the influxdb3-python
            # library [1]. Any migration should include tests.
            #
            # [1] https://github.com/InfluxCommunity/influxdb3-python
            warnings.warn(
                'Support for InfluxDB requires the influxdb library which '
                'only supports InfluxDB 1.x and is deprecated. As a result, '
                'influxdb is also deprecated and will be removed in a future '
                'release.',
                os_warnings.RemovedInSDK60Warning,
            )

        self._collector_registry = collector_registry

        self._service_type_manager = os_service_types.ServiceTypes()

    def __getattr__(self, key: str) -> Any:
        """Return arbitrary attributes.

        This method accesses config via __dict__ to avoid infinite recursion
        during copy.deepcopy(). When deepcopy creates a new instance without
        calling __init__, self.config doesn't exist, and accessing it directly
        would trigger __getattr__ again, causing a RecursionError.

        Dunder methods must raise AttributeError so Python can find inherited
        implementations (e.g., __reduce_ex__ from object for pickling/copying).
        """
        # Don't handle dunder methods - let Python find inherited impls
        if key.startswith('__') and key.endswith('__'):
            raise AttributeError(key)

        # Use __dict__.get() to safely check if config exists
        config = self.__dict__.get('config')
        if config is None:
            raise AttributeError(key)

        if key.startswith('os_'):
            key = key[3:]

        if key in [attr.replace('-', '_') for attr in config]:
            return config[key]
        else:
            return None

    def __iter__(self) -> Iterator[Any]:
        return self.config.__iter__()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CloudRegion):
            return NotImplemented

        return self.name == other.name and self.config == other.config

    def __ne__(self, other: object) -> bool:
        return not self == other

    @property
    def name(self) -> str:
        if self._name is not None:
            return self._name

        auth = self.get_session().auth
        # not all auth plugins are identity plugins
        if (
            auth
            and isinstance(auth, ks_identity_base.BaseIdentityPlugin)
            and auth.auth_url
        ):
            name = parse.urlparse(auth.auth_url).hostname or ''
        else:
            name = self._app_name or ''

        self._name = name

        return name

    @property
    def full_name(self) -> str:
        """Return a string that can be used as an identifier.

        Always returns a valid string. It will have name and region_name
        or just one of the two if only one is set, or else 'unknown'.
        """
        region_name = self.get_region_name()
        if self.name and region_name:
            return ":".join([self.name, region_name])
        elif self.name and not region_name:
            return self.name
        elif not self.name and region_name:
            return region_name
        else:
            return 'unknown'

    def set_service_value(
        self, key: str, service_type: str, value: Any
    ) -> None:
        key = _make_key(key, service_type)
        self.config[key] = value

    def set_session_constructor(
        self, session_constructor: type[ks_session.Session]
    ) -> None:
        """Sets the Session constructor."""
        self._session_constructor = session_constructor

    def get_requests_verify_args(
        self,
    ) -> tuple[bool | str | None, str | tuple[str, str] | None]:
        """Return the verify and cert values for the requests library."""
        insecure = self.config.get('insecure', False)
        verify = self.config.get('verify', True)
        cacert = self.config.get('cacert')
        # Insecure is the most aggressive setting, so it wins
        if insecure:
            verify = False
        if verify and cacert:
            verify = os.path.expanduser(cacert)
        else:
            if cacert:
                warnings.warn(
                    f"You are specifying a cacert for the cloud "
                    f"{self.full_name} but also to ignore the host "
                    f"verification. The host SSL cert will not be verified.",
                    os_warnings.ConfigurationWarning,
                )

        cert = self.config.get('cert')
        if cert:
            cert = os.path.expanduser(cert)
            if self.config.get('key'):
                cert = (cert, os.path.expanduser(self.config['key']))
        return (verify, cert)

    def get_services(self) -> list[str]:
        """Return a list of service types we know something about."""
        services = []
        for key, val in self.config.items():
            if (
                key.endswith('api_version')
                or key.endswith('service_type')
                or key.endswith('service_name')
            ):
                services.append("_".join(key.split('_')[:-2]))
        return list(set(services))

    def get_enabled_services(self) -> set[str]:
        services = set()

        all_services = [
            k['service_type'] for k in self._service_type_manager.services
        ]
        all_services.extend(
            k[4:] for k in self.config.keys() if k.startswith('has_')
        )

        for srv in all_services:
            ep = self.get_endpoint_from_catalog(srv)
            if ep:
                services.add(srv.replace('-', '_'))

        return services

    def get_auth_args(self) -> dict[str, Any]:
        return cast(dict[str, Any], self.config.get('auth', {}))

    @overload
    def _get_config(
        self,
        key: str,
        service_type: str | None,
        default: _T,
        fallback_to_unprefixed: bool = False,
    ) -> _T: ...

    @overload
    def _get_config(
        self,
        key: str,
        service_type: str | None,
        default: None = None,
        fallback_to_unprefixed: bool = False,
    ) -> Any | None: ...

    def _get_config(
        self,
        key: str,
        service_type: str | None,
        default: _T | None = None,
        fallback_to_unprefixed: bool = False,
    ) -> _T | Any | None:
        """Get a config value for a service_type.

        Finds the config value for a key, looking first for it prefixed by
        the given service_type, then by any known aliases of that service_type.
        Finally, if fallback_to_unprefixed is True, a value will be looked
        for without a prefix to support the config values where a global
        default makes sense.

        For instance, ``_get_config('example', 'block-storage')`` would
        first look for ``block_storage_example``, then ``volumev3_example``,
        ``volumev2_example`` and ``volume_example``. If no value was found, it
        would look for ``example``.

        If none of that works, it returns the value in ``default``.
        """
        if service_type is None:
            value = self.config.get(key)
        else:
            for st in self._service_type_manager.get_all_types(service_type):
                _key = _make_key(key, st)
                value = self.config.get(_key)
                if value is not None:
                    key = _key
                    break
            else:
                if fallback_to_unprefixed:
                    value = self.config.get(key)

        if value is None:
            return default
        else:
            return value

    def _get_service_config(self, key: str, service_type: str) -> Any | None:
        config_dict = self.config.get(key)
        if not config_dict:
            return None

        if not isinstance(config_dict, dict):
            raise RuntimeError(
                f'invalid configuration for service type {service_type!r}'
            )

        for st in self._service_type_manager.get_all_types(service_type):
            if st in config_dict:
                return config_dict[st]

        return None

    def get_region_name(self, service_type: str | None = None) -> str | None:
        # If a region_name for the specific service_type is configured, use it;
        # else use the one configured for the CloudRegion as a whole.
        value = self._get_config(
            'region_name', service_type, fallback_to_unprefixed=True
        )
        return str(value) if value is not None else value

    def get_interface(
        self, service_type: str | None = None
    ) -> list[str] | str | None:
        value = self._get_config(
            'interface', service_type, fallback_to_unprefixed=True
        )
        if value is None:
            return value
        if isinstance(value, str):
            return value
        if isinstance(value, list) and all(
            {isinstance(x, str) or x is None for x in value}
        ):
            return value

        raise exceptions.ConfigException(
            f'interface should be str, list of str or None but is '
            f'{type(value)}'
        )

    def get_api_version(self, service_type: str) -> str | None:
        version = self._get_config('api_version', service_type)
        if not version:
            return None

        try:
            float(version)
        except ValueError:
            if 'latest' in version:
                warnings.warn(
                    "You have a configured API_VERSION with 'latest' in "
                    "it. In the context of openstacksdk this doesn't make "
                    "any sense.",
                    os_warnings.ConfigurationWarning,
                )
            return None

        return str(version)

    def get_default_microversion(self, service_type: str) -> str | None:
        value = self._get_config('default_microversion', service_type)
        return str(value) if value is not None else value

    def get_service_type(self, service_type: str) -> str:
        # People requesting 'volume' are doing so because os-client-config
        # let them. What they want is block-storage, not explicitly the
        # v1 of cinder. If someone actually wants v1, they'll have api_version
        # set to 1, in which case block-storage will still work properly.
        # Use service-types-manager to grab the official type name. _get_config
        # will still look for config by alias, but starting with the official
        # type will get us things in the right order.
        if self._service_type_manager.is_known(service_type):
            official_type = self._service_type_manager.get_service_type(
                service_type
            )
            if official_type is not None:
                service_type = official_type
        value = self._get_config(
            'service_type', service_type, default=service_type
        )
        return str(value) if value is not None else value

    def get_service_name(self, service_type: str) -> str | None:
        value = self._get_config('service_name', service_type)
        return str(value) if value is not None else value

    def get_endpoint(self, service_type: str) -> str | None:
        auth = self.config.get('auth', {})
        value = self._get_config('endpoint_override', service_type)
        if not value:
            value = self._get_config('endpoint', service_type)

        if not value and self.config.get('auth_type') == 'none':
            # If endpoint is given and we're using the none auth type,
            # then the endpoint value is the endpoint_override for every
            # service.
            value = auth.get('endpoint')

        if (
            not value
            and service_type == 'identity'
            and SCOPE_KEYS.isdisjoint(set(auth.keys()))
        ):
            # There are a small number of unscoped identity operations.
            # Specifically, looking up a list of projects/domains/system to
            # scope to.
            value = auth.get('auth_url')

        # Because of course. Seriously.
        # We have to override the Rackspace block-storage endpoint because
        # only v1 is in the catalog but the service actually does support
        # v2. But the endpoint needs the project_id.
        official_type = self._service_type_manager.get_service_type(
            service_type
        )
        if official_type is not None:
            service_type = official_type
        if (
            value
            and self.config.get('profile') == 'rackspace'
            and service_type == 'block-storage'
        ):
            value = value + auth.get('project_id')

        return str(value) if value else None

    def get_endpoint_from_catalog(
        self,
        service_type: str,
        interface: list[str] | str | None = None,
        region_name: str | None = None,
    ) -> str | None:
        """Return the endpoint for a given service as found in the catalog.

        For values respecting endpoint overrides, see
        :meth:`~openstack.connection.Connection.endpoint_for`

        :param service_type: Service Type of the endpoint to search for.
        :param interface:
            Interface of the endpoint to search for. Optional, defaults to
            the configured value for interface for this Connection.
        :param region_name:
            Region Name of the endpoint to search for. Optional, defaults to
            the configured value for region_name for this Connection.

        :returns: The endpoint of the service, or None if not found.
        """
        interface = interface or self.get_interface(service_type)
        region_name = region_name or self.get_region_name(service_type)
        session = self.get_session()

        auth = session.auth
        if not isinstance(auth, ks_identity_base.BaseIdentityPlugin):
            return None

        catalog = auth.get_access(session).service_catalog
        try:
            return catalog.url_for(
                service_type=service_type,
                interface=interface,
                region_name=region_name,
            )
        except (keystoneauth1.exceptions.catalog.EndpointNotFound, ValueError):
            return None

    def get_connect_retries(self, service_type: str) -> int | None:
        value = self._get_config(
            'connect_retries', service_type, fallback_to_unprefixed=True
        )
        return int(value) if value is not None else value

    def get_connect_retry_delay(self, service_type: str) -> float | None:
        value = self._get_config(
            'connect_retry_delay', service_type, fallback_to_unprefixed=True
        )
        return float(value) if value is not None else value

    def get_status_code_retries(self, service_type: str) -> int | None:
        value = self._get_config(
            'status_code_retries', service_type, fallback_to_unprefixed=True
        )
        return int(value) if value is not None else value

    @property
    def prefer_ipv6(self) -> bool:
        return not self._force_ipv4

    @property
    def force_ipv4(self) -> bool:
        return self._force_ipv4

    def get_auth(self) -> plugin.BaseAuthPlugin | None:
        """Return a keystoneauth plugin from the auth credentials."""
        return self._auth

    def skip_auth_cache(self) -> bool:
        return not keyring or not self._auth or not self._cache_auth

    def load_auth_from_cache(self) -> None:
        if self.skip_auth_cache():
            return

        assert self._auth is not None  # narrow type

        cache_id = self._auth.get_cache_id()

        # skip if the plugin does not support caching
        if not cache_id:
            return

        try:
            state = keyring.get_password('openstacksdk', cache_id)
        except RuntimeError:  # the fail backend raises this
            state = None

        if not state:
            self.log.debug('Failed to fetch auth from keyring')
            return

        self.log.debug('Reusing authentication from keyring')
        self._auth.set_auth_state(state)

    def set_auth_cache(self) -> None:
        if self.skip_auth_cache():
            return

        assert self._auth is not None  # narrow type

        cache_id = self._auth.get_cache_id()
        # NOTE(stephenfin): The actual return type of this is a serialized JSON
        # object
        state = cast(str, self._auth.get_auth_state())

        try:
            if cache_id and state:
                # NOTE: under some conditions the method may be invoked when
                # auth is empty. This may lead to exception in the keyring lib,
                # thus do nothing.
                keyring.set_password('openstacksdk', cache_id, state)
        except RuntimeError:  # the fail backend raises this
            self.log.debug('Failed to set auth into keyring')

    def insert_user_agent(self) -> None:
        """Set sdk information into the user agent of the Session.

        .. warning::
            This method is here to be used by os-client-config. It exists
            as a hook point so that os-client-config can provice backwards
            compatibility and still be in the User Agent for people using
            os-client-config directly.

        Normal consumers of SDK should use app_name and app_version. However,
        if someone else writes a subclass of
        :class:`~openstack.config.cloud_region.CloudRegion` it may be
        desirable.
        """
        if not self._keystone_session:
            return

        self._keystone_session.additional_user_agent.append(
            ('openstacksdk', importlib_metadata.version('openstacksdk'))
        )

    def get_session(self) -> ks_session.Session:
        """Return a keystoneauth session based on the auth credentials."""
        if self._keystone_session is not None:
            return self._keystone_session

        if not self._auth:
            raise exceptions.ConfigException("Problem with auth parameters")

        verify, cert = self.get_requests_verify_args()

        warnings.filterwarnings(
            'ignore', category=urllib3.exceptions.InsecurePlatformWarning
        )
        # Turn off urllib3 warnings about insecure certs if we have
        # explicitly configured requests to tell it we do not want
        # cert verification
        if not verify:
            self.log.debug(
                f"Turning off SSL warnings for {self.full_name} "
                f"since verify=False"
            )
            warnings.filterwarnings(
                'ignore',
                category=urllib3.exceptions.InsecureRequestWarning,
            )

        self._keystone_session = self._session_constructor(
            auth=self._auth,
            verify=verify,
            cert=cert,
            timeout=self.config.get('api_timeout'),
            collect_timing=bool(self.config.get('timing')),
            discovery_cache=self._discovery_cache,
        )
        self.insert_user_agent()
        # Using old keystoneauth with new os-client-config fails if
        # we pass in app_name and app_version. Those are not essential,
        # nor a reason to bump our minimum, so just test for the session
        # having the attribute post creation and set them then.
        if hasattr(self._keystone_session, 'app_name'):
            self._keystone_session.app_name = self._app_name
        if hasattr(self._keystone_session, 'app_version'):
            self._keystone_session.app_version = self._app_version

        return self._keystone_session

    def get_service_catalog(
        self,
    ) -> ks_service_catalog.ServiceCatalog | None:
        """Helper method to grab the service catalog."""
        # not all auth plugins are identity plugins
        if not isinstance(self._auth, ks_identity_base.BaseIdentityPlugin):
            return None
        return self._auth.get_access(self.get_session()).service_catalog

    def _get_version_request(
        self, service_type: str, version: str | None
    ) -> _util.VersionRequest:
        """Translate OCC version args to those needed by ksa adapter.

        If no version is requested explicitly and we have a configured version,
        set the version parameter and let ksa deal with expanding that to
        min=ver.0, max=ver.latest.

        If version is set, pass it through.

        If version is not set and we don't have a configured version, default
        to latest.

        If version is set, contains a '.', and default_microversion is not
        set, also pass it as a default microversion.
        """
        version_request = _util.VersionRequest()
        if version == 'latest':
            version_request.max_api_version = 'latest'
            return version_request

        if not version:
            version = self.get_api_version(service_type)

        # Octavia doens't have a version discovery document. Hard-code an
        # exception to this logic for now.
        if not version and service_type not in ('load-balancer',):
            version_request.max_api_version = 'latest'
        else:
            version_request.version = version

        default_microversion = self.get_default_microversion(service_type)
        implied_microversion = _get_implied_microversion(version)
        if (
            implied_microversion
            and default_microversion
            and implied_microversion != default_microversion
        ):
            raise exceptions.ConfigException(
                f"default_microversion of {default_microversion} was given "
                f"for {service_type}, but api_version looks like a "
                f"microversion as well. Please set api_version to just the "
                f"desired major version, or omit default_microversion"
            )
        if implied_microversion:
            assert version is not None  # type narrow
            default_microversion = implied_microversion
            # If we're inferring a microversion, don't pass the whole
            # string in as api_version, since that tells keystoneauth
            # we're looking for a major api version.
            version_request.version = version[0]

        version_request.default_microversion = default_microversion

        return version_request

    def get_all_version_data(
        self, service_type: str
    ) -> list[discover.VersionData]:
        """Retrieve version data for the given service.

        :param service_type: The service to fetch version data for.
        :returns: A `~keystoneauth1.discover.VersionData` object containing the
            version data for the requested service.
        """
        # Seriously. Don't think about the existential crisis
        # that is the next line. You'll wind up in cthulhu's lair.
        service_type = self.get_service_type(service_type)
        region_name = self.get_region_name(service_type)
        assert region_name is not None  # narrow type
        interface = self.get_interface(service_type)
        assert interface is not None  # narrow type

        interfaces = interface if isinstance(interface, list) else [interface]

        versions = self.get_session().get_all_version_data(
            service_type=service_type,
            interface=interface,
            region_name=region_name,
        )

        region_versions = versions.get(region_name, {})
        for interface in interfaces:
            interface_versions = region_versions.get(interface, {})
            service_version_data = interface_versions.get(service_type)
            if service_version_data is not None:
                return service_version_data

        return []

    def _get_endpoint_from_catalog(
        self,
        service_type: str,
        constructor: type[proxy.Proxy],
    ) -> str:
        adapter = constructor(
            session=self.get_session(),
            service_type=self.get_service_type(service_type),
            service_name=self.get_service_name(service_type),
            # https://review.opendev.org/c/openstack/keystoneauth/+/951183
            interface=self.get_interface(service_type),  # type: ignore
            region_name=self.get_region_name(service_type),
        )
        endpoint = adapter.get_endpoint()
        assert endpoint is not None  # narrow type
        return endpoint

    def _get_hardcoded_endpoint(
        self, service_type: str, constructor: type[proxy.Proxy]
    ) -> str:
        endpoint = self._get_endpoint_from_catalog(service_type, constructor)
        if not endpoint.rstrip().rsplit('/')[-1] == 'v2.0':
            if not endpoint.endswith('/'):
                endpoint += '/'
            endpoint = parse.urljoin(endpoint, 'v2.0')
        return endpoint

    @overload
    def get_session_client(
        self,
        service_type: str,
        version: str | None = None,
        constructor: None = None,
        **kwargs: Any,
    ) -> proxy.Proxy: ...

    @overload
    def get_session_client(
        self,
        service_type: str,
        version: str | None = None,
        constructor: type[proxy.ProxyT] = ...,
        **kwargs: Any,
    ) -> proxy.ProxyT: ...

    def get_session_client(
        self,
        service_type: str,
        version: str | None = None,
        constructor: type[proxy.Proxy] | None = None,
        **kwargs: Any,
    ) -> proxy.Proxy:
        """Return a prepped keystoneauth Adapter for a given service.

        This is useful for making direct requests calls against a
        'mounted' endpoint. That is, if you do:

          client = get_session_client('compute')

        then you can do:

          client.get('/flavors')

        and it will work like you think.
        """
        if constructor is None:
            constructor = proxy.Proxy

        version_request = self._get_version_request(service_type, version)

        kwargs.setdefault('region_name', self.get_region_name(service_type))
        kwargs.setdefault(
            'connect_retries', self.get_connect_retries(service_type)
        )
        kwargs.setdefault(
            'connect_retry_delay', self.get_connect_retry_delay(service_type)
        )
        kwargs.setdefault(
            'status_code_retries', self.get_status_code_retries(service_type)
        )
        kwargs.setdefault('statsd_prefix', self.get_statsd_prefix())
        kwargs.setdefault('statsd_client', self.get_statsd_client())
        kwargs.setdefault('prometheus_counter', self.get_prometheus_counter())
        kwargs.setdefault(
            'prometheus_histogram', self.get_prometheus_histogram()
        )
        kwargs.setdefault('influxdb_config', self._influxdb_config)
        kwargs.setdefault('influxdb_client', self.get_influxdb_client())
        endpoint_override = self.get_endpoint(service_type)
        version = version_request.version
        min_api_version = (
            kwargs.pop('min_version', None) or version_request.min_api_version
        )
        max_api_version = (
            kwargs.pop('max_version', None) or version_request.max_api_version
        )

        # Older neutron has inaccessible discovery document. Nobody noticed
        # because neutronclient hard-codes an append of v2.0. YAY!
        # Also, older octavia has a similar issue.
        if service_type in ('network', 'load-balancer'):
            version = None
            min_api_version = None
            max_api_version = None
            if endpoint_override is None:
                endpoint_override = self._get_hardcoded_endpoint(
                    service_type, constructor
                )

        client = constructor(
            session=self.get_session(),
            service_type=self.get_service_type(service_type),
            service_name=self.get_service_name(service_type),
            # https://review.opendev.org/c/openstack/keystoneauth/+/951183
            interface=self.get_interface(service_type),  # type: ignore
            version=version,
            min_version=min_api_version,
            max_version=max_api_version,
            endpoint_override=endpoint_override,
            default_microversion=version_request.default_microversion,
            rate_limit=self.get_rate_limit(service_type),
            concurrency=self.get_concurrency(service_type),
            **kwargs,
        )
        if version_request.default_microversion:
            default_microversion = version_request.default_microversion
            info = client.get_endpoint_data()
            if info and not discover.version_between(
                info.min_microversion,
                info.max_microversion,
                default_microversion,
            ):
                if self.get_default_microversion(service_type):
                    raise exceptions.ConfigException(
                        "A default microversion for service {service_type} of "
                        "{default_microversion} was requested, but the cloud "
                        "only supports a minimum of {min_microversion} and "
                        "a maximum of {max_microversion}.".format(
                            service_type=service_type,
                            default_microversion=default_microversion,
                            min_microversion=discover.version_to_string(
                                info.min_microversion or (0,)
                            ),
                            max_microversion=discover.version_to_string(
                                info.max_microversion or (0,)
                            ),
                        )
                    )
                else:
                    raise exceptions.ConfigException(
                        "A default microversion for service {service_type} of "
                        "{default_microversion} was requested, but the cloud "
                        "only supports a minimum of {min_microversion} and "
                        "a maximum of {max_microversion}. The default "
                        "microversion was set because a microversion "
                        "formatted version string, '{api_version}', was "
                        "passed for the api_version of the service. If it "
                        "was not intended to set a default microversion "
                        "please remove anything other than an integer major "
                        "version from the version setting for "
                        "the service.".format(
                            service_type=service_type,
                            api_version=self.get_api_version(service_type),
                            default_microversion=default_microversion,
                            min_microversion=discover.version_to_string(
                                info.min_microversion or (0,)
                            ),
                            max_microversion=discover.version_to_string(
                                info.max_microversion or (0,)
                            ),
                        )
                    )
        return client

    def get_session_endpoint(
        self,
        service_type: str,
        min_version: str | None = None,
        max_version: str | None = None,
    ) -> str | None:
        """Return the endpoint from config or the catalog.

        If a configuration lists an explicit endpoint for a service,
        return that. Otherwise, fetch the service catalog from the
        keystone session and return the appropriate endpoint.

        :param service_type: Official service type of service
        """

        override_endpoint = self.get_endpoint(service_type)
        if override_endpoint:
            return override_endpoint

        region_name = self.get_region_name(service_type)
        service_name = self.get_service_name(service_type)
        interface = self.get_interface(service_type)
        session = self.get_session()
        try:
            # Return the highest version we find that matches
            # the request
            endpoint = session.get_endpoint(
                service_type=service_type,
                region_name=region_name,
                interface=interface,
                service_name=service_name,
                min_version=min_version,
                max_version=max_version,
            )
        except keystoneauth1.exceptions.catalog.EndpointNotFound:
            endpoint = None
        if not endpoint:
            self.log.warning(
                "Keystone catalog entry not found ("
                "service_type=%s,service_name=%s,"
                "interface=%s,region_name=%s)",
                service_type,
                service_name,
                interface,
                region_name,
            )
        return endpoint

    def get_cache_expiration_time(self) -> int:
        # TODO(mordred) We should be validating/transforming this on input
        return int(self._cache_expiration_time)

    def get_cache_path(self) -> str | None:
        return self._cache_path

    def get_cache_class(self) -> str:
        return self._cache_class

    def get_cache_arguments(self) -> dict[str, Any] | None:
        return copy.deepcopy(self._cache_arguments)

    def get_cache_expirations(self) -> dict[str, int]:
        return copy.deepcopy(self._cache_expirations)

    def get_cache_resource_expiration(
        self, resource: str, default: float | None = None
    ) -> float | None:
        """Get expiration time for a resource

        :param resource: Name of the resource type
        :param default: Default value to return if not found (optional,
                        defaults to None)

        :returns: Expiration time for the resource type as float or default
        """
        if resource not in self._cache_expirations:
            return default
        return float(self._cache_expirations[resource])

    def requires_floating_ip(self) -> bool | None:
        """Return whether or not this cloud requires floating ips.


        :returns: True or False if know, None if discovery is needed.
                  If requires_floating_ip is not configured but the cloud is
                  known to not provide floating ips, will return False.
        """
        if self.config['floating_ip_source'] == "None":
            return False
        requires_floating_ip = self.config.get('requires_floating_ip')
        if requires_floating_ip is None:
            return None
        return bool(requires_floating_ip)

    def get_external_networks(self) -> list[str]:
        """Get list of network names for external networks."""
        return [
            net['name']
            for net in self.config.get('networks', [])
            if net['routes_externally']
        ]

    def get_external_ipv4_networks(self) -> list[str]:
        """Get list of network names for external IPv4 networks."""
        return [
            str(net['name'])
            for net in self.config.get('networks', [])
            if net['routes_ipv4_externally']
        ]

    def get_external_ipv6_networks(self) -> list[str]:
        """Get list of network names for external IPv6 networks."""
        return [
            str(net['name'])
            for net in self.config.get('networks', [])
            if net['routes_ipv6_externally']
        ]

    def get_internal_networks(self) -> list[str]:
        """Get list of network names for internal networks."""
        return [
            str(net['name'])
            for net in self.config.get('networks', [])
            if not net['routes_externally']
        ]

    def get_internal_ipv4_networks(self) -> list[str]:
        """Get list of network names for internal IPv4 networks."""
        return [
            str(net['name'])
            for net in self.config.get('networks', [])
            if not net['routes_ipv4_externally']
        ]

    def get_internal_ipv6_networks(self) -> list[str]:
        """Get list of network names for internal IPv6 networks."""
        return [
            str(net['name'])
            for net in self.config.get('networks', [])
            if not net['routes_ipv6_externally']
        ]

    def get_default_network(self) -> str | None:
        """Get network used for default interactions."""
        for net in self.config.get('networks', []):
            if net['default_interface']:
                return str(net['name'])
        return None

    def get_nat_destination(self) -> str | None:
        """Get network used for NAT destination."""
        for net in self.config.get('networks', []):
            if net['nat_destination']:
                return str(net['name'])
        return None

    def get_nat_source(self) -> str | None:
        """Get network used for NAT source."""
        for net in self.config.get('networks', []):
            if net.get('nat_source'):
                return str(net['name'])
        return None

    def _get_extra_config(
        self,
        key: str | None,
        defaults: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Fetch an arbitrary extra chunk of config, laying in defaults.

        :param string key: name of the config section to fetch
        :param dict defaults: (optional) default values to merge under the
                                         found config
        """
        defaults = _util.normalize_keys(defaults or {})
        if not key:
            return defaults or {}
        return _util.merge_clouds(
            defaults, _util.normalize_keys(self._extra_config.get(key, {}))
        )

    def get_client_config(
        self,
        name: str | None = None,
        defaults: dict[str, Any] | None = None,
    ) -> dict[str, Any] | None:
        """Get config settings for a named client.

        Settings will also be looked for in a section called 'client'.
        If settings are found in both, they will be merged with the settings
        from the named section winning over the settings from client section,
        and both winning over provided defaults.

        :param string name:
            Name of the config section to look for.
        :param dict defaults:
            Default settings to use.

        :returns:
            A dict containing merged settings from the named section, the
            client section and the defaults.
        """
        return self._get_extra_config(
            name, self._get_extra_config('client', defaults)
        )

    def get_password_callback(self) -> _PasswordCallback | None:
        return self._password_callback

    def get_rate_limit(self, service_type: str) -> float | None:
        return self._get_service_config(
            'rate_limit', service_type=service_type
        )

    def get_concurrency(self, service_type: str) -> int | None:
        return self._get_service_config(
            'concurrency', service_type=service_type
        )

    def get_statsd_client(
        self,
    ) -> Optional['statsd_client.StatsClientBase']:
        if not statsd_client:
            if self._statsd_host:
                self.log.warning(
                    'StatsD python library is not available. '
                    'Reporting disabled'
                )
            return None
        statsd_args = {}
        if self._statsd_host:
            statsd_args['host'] = self._statsd_host
        if self._statsd_port:
            statsd_args['port'] = self._statsd_port
        if statsd_args:
            try:
                return statsd_client.StatsClient(**statsd_args)
            except Exception:
                self.log.warning('Cannot establish connection to statsd')
                return None
        else:
            return None

    def get_statsd_prefix(self) -> str:
        return self._statsd_prefix or 'openstack.api'

    def get_prometheus_registry(
        self,
    ) -> Optional['prometheus_client.CollectorRegistry']:
        return self._collector_registry

    def get_prometheus_histogram(
        self,
    ) -> Optional['prometheus_client.Histogram']:
        registry = self.get_prometheus_registry()
        if not registry or not prometheus_client:
            return None
        # We have to hide a reference to the histogram on the registry
        # object, because it's collectors must be singletons for a given
        # registry but register at creation time.
        hist = getattr(registry, '_openstacksdk_histogram', None)
        if not hist:
            hist = prometheus_client.Histogram(
                'openstack_http_response_time',
                'Time taken for an http response to an OpenStack service',
                labelnames=[
                    'method',
                    'endpoint',
                    'service_type',
                    'status_code',
                ],
                registry=registry,
            )
            setattr(registry, '_openstacksdk_histogram', hist)
        return hist

    def get_prometheus_counter(
        self,
    ) -> Optional['prometheus_client.Counter']:
        registry = self.get_prometheus_registry()
        if not registry or not prometheus_client:
            return None
        counter = getattr(registry, '_openstacksdk_counter', None)
        if not counter:
            counter = prometheus_client.Counter(
                'openstack_http_requests',
                'Number of HTTP requests made to an OpenStack service',
                labelnames=[
                    'method',
                    'endpoint',
                    'service_type',
                    'status_code',
                ],
                registry=registry,
            )
            setattr(registry, '_openstacksdk_counter', counter)
        return counter

    def has_service(self, service_type: str) -> bool:
        service_type = service_type.lower().replace('-', '_')
        key = f'has_{service_type}'
        value = self.config.get(
            key, self._service_type_manager.is_official(service_type)
        )
        assert isinstance(value, bool)
        return value

    def disable_service(
        self, service_type: str, reason: str | None = None
    ) -> None:
        _disable_service(self.config, service_type, reason=reason)

    def enable_service(self, service_type: str) -> None:
        service_type = service_type.lower().replace('-', '_')
        key = f'has_{service_type}'
        self.config[key] = True

    def get_disabled_reason(self, service_type: str) -> str | None:
        service_type = service_type.lower().replace('-', '_')
        d_key = _make_key('disabled_reason', service_type)
        return self.config.get(d_key)

    def get_influxdb_client(
        self,
    ) -> Optional['influxdb_client.InfluxDBClient']:
        influx_args: dict[str, Any] = {}
        if not self._influxdb_config:
            return None

        warnings.warn(
            'Support for InfluxDB requires the influxdb library which '
            'only supports InfluxDB 1.x and is deprecated. As a result, '
            'influxdb is also deprecated and will be removed in a future '
            'release.',
            os_warnings.RemovedInSDK60Warning,
        )

        use_udp = bool(self._influxdb_config.get('use_udp', False))
        port = self._influxdb_config.get('port')
        if use_udp:
            influx_args['use_udp'] = True
        if 'port' in self._influxdb_config:
            if use_udp:
                influx_args['udp_port'] = port
            else:
                influx_args['port'] = port
        for key in ['host', 'username', 'password', 'database', 'timeout']:
            if key in self._influxdb_config:
                influx_args[key] = self._influxdb_config[key]
        if influxdb_client and influx_args:
            try:
                return influxdb_client.InfluxDBClient(**influx_args)
            except Exception:
                self.log.warning('Cannot establish connection to InfluxDB')
        else:
            self.log.warning(
                'InfluxDB configuration is present, '
                'but no client library is found.'
            )
        return None



# Copyright 2026 Red Hat Inc.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstack.image.v2 import image_location
from openstack.tests.unit import base

EXAMPLE = {
    'image_id': '2',
    'url': 'http://spam.com/',
    'validation_data': {'os_hash_algo': 'sha512'},
    'metadata': {},
}


class TestImageLocation(base.TestCase):
    def test_basic(self):
        sot = image_location.ImageLocation()
        self.assertEqual('/images/%(image_id)s/locations', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = image_location.ImageLocation(**EXAMPLE)
        self.assertEqual(EXAMPLE['image_id'], sot.image_id)
        self.assertEqual(EXAMPLE['url'], sot.url)
        self.assertEqual(EXAMPLE['validation_data'], sot.validation_data)
        self.assertEqual(EXAMPLE['metadata'], sot.metadata)

    def test_prepare_request_body_includes_validation_data(self):
        sot = image_location.ImageLocation(
            image_id='test_id',
            url='http://spam.com/',
            validation_data={'os_hash_algo': 'sha512'},
        )
        body = sot._prepare_request_body(prepend_key=False)
        self.assertEqual(
            {
                'url': 'http://spam.com/',
                'validation_data': {'os_hash_algo': 'sha512'},
            },
            body,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.key_manager.v1 import quota as _quota
from openstack.tests.functional import base


class TestQuota(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()
        self.require_service('key-manager')

    def test_quota_get(self):
        """Test getting current project's effective quotas."""
        quota = self.user_cloud.key_manager.get_quota()

        self.assertIsInstance(quota, _quota.Quota)

        # Verify quota attributes exist and are integers
        self.assertIsInstance(quota.secrets, int)
        self.assertIsInstance(quota.orders, int)
        self.assertIsInstance(quota.containers, int)
        self.assertIsInstance(quota.consumers, int)
        self.assertIsInstance(quota.cas, int)

        # Verify quotas are reasonable values (not None or negative)
        self.assertGreaterEqual(quota.secrets, -1)  # -1 means unlimited
        self.assertGreaterEqual(quota.orders, -1)
        self.assertGreaterEqual(quota.containers, -1)
        self.assertGreaterEqual(quota.consumers, -1)
        self.assertGreaterEqual(quota.cas, -1)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Delete resources from the Key Manager service.

For a full guide see
https://docs.openstack.org/openstacksdk/latest/user/guides/key_manager.html
"""


def delete_secret(conn, name_or_id):
    print(f"Delete Secret: {name_or_id}")

    secret = conn.key_manager.find_secret(name_or_id)

    if secret:
        conn.key_manager.delete_secret(secret)
    else:
        print("Secret not found")


def delete_secret_consumer(conn, secret):
    print("Delete a secret consumer:")

    conn.key_manager.delete_secret_consumer(
        secret=secret,
        service="service",
        resource_type="resource_type",
        resource_id="resource-id",
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional.shared_file_system.v2 import base


class ShareAccessRuleTest(base.BaseSharedFileSystemTest):
    def setUp(self):
        super().setUp()

        self.SHARE_NAME = self.getUniqueString()
        mys = self.create_share(
            name=self.SHARE_NAME,
            size=2,
            share_type="dhss_false",
            share_protocol='NFS',
            description=None,
        )
        self.user_cloud.shared_file_system.wait_for_status(
            mys,
            status='available',
            failures=['error'],
            interval=5,
            wait=self._wait_for_timeout,
        )
        self.assertIsNotNone(mys)
        self.assertIsNotNone(mys.id)
        self.SHARE_ID = mys.id
        self.SHARE = mys
        access_rule = self.user_cloud.share.create_access_rule(
            self.SHARE_ID,
            access_level="rw",
            access_type="ip",
            access_to="0.0.0.0/0",
        )
        self.ACCESS_ID = access_rule.id
        self.RESOURCE_KEY = access_rule.resource_key

    def tearDown(self):
        self.user_cloud.share.delete_access_rule(
            self.ACCESS_ID, self.SHARE_ID, ignore_missing=True
        )
        super().tearDown()

    def test_get_access_rule(self):
        sot = self.user_cloud.shared_file_system.get_access_rule(
            self.ACCESS_ID
        )
        self.assertEqual(self.ACCESS_ID, sot.id)

    def test_list_access_rules(self):
        rules = self.user_cloud.shared_file_system.access_rules(
            self.SHARE, details=True
        )
        self.assertGreater(len(list(rules)), 0)
        for rule in rules:
            for attribute in (
                'id',
                'created_at',
                'updated_at',
                'access_level',
                'access_type',
                'access_to',
                'share_id',
                'access_key',
                'metadata',
            ):
                self.assertTrue(hasattr(rule, attribute))

    def test_create_delete_access_rule_with_locks(self):
        access_rule = self.user_cloud.share.create_access_rule(
            self.SHARE_ID,
            access_level="rw",
            access_type="ip",
            access_to="203.0.113.10",
            lock_deletion=True,
            lock_visibility=True,
        )

        self.user_cloud.share.delete_access_rule(
            access_rule['id'], self.SHARE_ID, unrestrict=True
        )
        self.user_cloud.shared_file_system.wait_for_delete(access_rule)



# Licensed under the Apache License, Version 2.0 (the "License"); you may

# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
from typing import Any, Self
import urllib.parse

from keystoneauth1 import adapter

from openstack import exceptions
from openstack import resource
from openstack import utils


class Info(resource.Resource):
    base_path = "/info"

    allow_fetch = True

    _query_mapping = resource.QueryParameters(
        'swiftinfo_sig', 'swiftinfo_expires'
    )

    # Properties
    bulk_delete = resource.Body("bulk_delete", type=dict)
    swift = resource.Body("swift", type=dict)
    slo = resource.Body("slo", type=dict)
    staticweb = resource.Body("staticweb", type=dict)
    tempurl = resource.Body("tempurl", type=dict)

    # The endpoint in the catalog has version and project-id in it
    # To get capabilities, we have to disassemble and reassemble the URL
    # to append 'info'
    # This logic is taken from swiftclient
    def _get_info_url(self, url: str) -> str:
        URI_PATTERN_VERSION = re.compile(r'\/v\d+\.?\d*(\/.*)?')
        scheme, netloc, path, params, query, fragment = urllib.parse.urlparse(
            url
        )
        if URI_PATTERN_VERSION.search(path):
            path = URI_PATTERN_VERSION.sub('/info', path)
        else:
            path = utils.urljoin(path, 'info')

        return urllib.parse.urlunparse(
            (scheme, netloc, path, params, query, fragment)
        )

    def fetch(
        self,
        session: adapter.Adapter,
        requires_id: bool = False,
        base_path: str | None = None,
        error_message: str | None = None,
        skip_cache: bool = False,
        **kwargs: Any,
    ) -> Self:
        """Get a remote resource based on this instance.

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param boolean requires_id: A boolean indicating whether resource ID
                                    should be part of the requested URI.
        :param str base_path: Base part of the URI for fetching resources, if
                              different from
                              :data:`~openstack.resource.Resource.base_path`.
        :param str error_message: An Error message to be returned if
                                  requested object does not exist.
        :return: This :class:`Resource` instance.
        :raises: :exc:`~openstack.exceptions.MethodNotSupported` if
                 :data:`Resource.allow_fetch` is not set to ``True``.
        :raises: :exc:`~openstack.exceptions.NotFoundException` if
                 the resource was not found.
        """
        if not self.allow_fetch:
            raise exceptions.MethodNotSupported(self, "fetch")

        session = self._get_session(session)
        info_url = self._get_info_url(session.get_endpoint() or '')

        microversion = self._get_microversion(session)
        response = session.get(info_url, microversion=microversion)

        self.microversion = microversion
        self._translate_response(response, error_message=error_message)
        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Self

from keystoneauth1 import adapter

from openstack import exceptions
from openstack import resource
from openstack import utils


class ConsistencyGroup(resource.Resource):
    resource_key = "consistencygroup"
    resources_key = "consistencygroups"
    base_path = "/consistencygroups"

    _query_mapping = resource.QueryParameters(
        "limit",
        "marker",
        "offset",
        "sort_dir",
        "sort_key",
        "sort",
    )

    # capabilities
    allow_fetch = True
    allow_create = True
    allow_delete = True
    allow_commit = True
    allow_list = True

    # Properties
    #: The availability zone.
    availability_zone = resource.Body("availability_zone")
    #: The date and time when this resource was created.
    created_at = resource.Body("created_at")
    #: The description.
    description = resource.Body("description")
    #: The name.
    name = resource.Body("name")
    #: The status.
    status = resource.Body("status")
    #: The volume types.
    volume_types = resource.Body("volume_types", type=list)
    #: Comma-separated list of volume UUIDs to add to the consistency group.
    #: Used in update requests only.
    add_volumes = resource.Body("add_volumes")
    #: Comma-separated list of volume UUIDs to remove from the consistency
    #: group. Used in update requests only.
    remove_volumes = resource.Body("remove_volumes")

    @classmethod
    def create_from_source(
        cls,
        session: adapter.Adapter,
        *,
        consistency_group_snapshot_id: str | None = None,
        consistency_group_id: str | None = None,
        name: str | None = None,
        description: str | None = None,
    ) -> Self:
        """Creates a new group from source.

        :param session: The session to use for making this request
        :param consistency_group_snapshot_id: The ID of a consistency group
            snapshot to create the new consistency group from. Either this or
            consistency_group_id must be provided
        :param consistency_group_id: The ID of a consistency group to create
            the new consistency group from. Either this or
            consistency_group_snapshot_id must be provided
        :param name: The name to assign to the new consistency group
        :param description: The description to set on the new consistency
            group
        :returns: The created consistency group
        """
        session = cls._get_session(session)
        microversion = cls._get_microversion(session)
        url = utils.urljoin(cls.base_path, 'create_from_src')
        body = {
            'consistencygroup-from-src': {
                'name': name,
                'description': description,
                'cgsnapshot_id': consistency_group_snapshot_id,
                'source_cgid': consistency_group_id,
            }
        }
        response = session.post(url, json=body, microversion=microversion)
        exceptions.raise_from_response(response)

        cg = cls()
        cg._translate_response(response=response)
        return cg

    def delete(
        self,
        session: adapter.Adapter,
        error_message: str | None = None,
        *,
        microversion: str | None = None,
        base_path: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> Self:
        # consistency group delete is done via a POST, not a DELETE :(
        url = utils.urljoin(self.base_path, self.id, 'delete')
        body: dict[str, Any] = {'consistency_group': {}}
        if params and 'force' in params:
            body['consistency_group']['force'] = params['force']
        response = session.post(url, json=body, microversion=microversion)

        self._translate_response(
            response, has_body=False, error_message=error_message
        )
        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from examples.identity import list as identity_list

from openstack.tests.functional import base


class TestIdentity(base.BaseFunctionalTest):
    """Test the identity examples

    The purpose of these tests is to ensure the examples run successfully.
    """

    def test_identity(self):
        identity_list.list_users(self.operator_cloud)
        identity_list.list_credentials(self.operator_cloud)
        identity_list.list_projects(self.operator_cloud)
        identity_list.list_domains(self.operator_cloud)
        identity_list.list_groups(self.operator_cloud)
        identity_list.list_services(self.operator_cloud)
        identity_list.list_endpoints(self.operator_cloud)
        identity_list.list_regions(self.operator_cloud)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_server_set_metadata
----------------------------------

Tests for the `set_server_metadata` command.
"""

import uuid

from openstack import exceptions
from openstack.tests.unit import base
from openstack.tests.unit.cloud import fakes


class TestServerSetMetadata(base.TestCase):
    def setUp(self):
        super().setUp()
        self.server_id = str(uuid.uuid4())
        self.server_name = self.getUniqueString('name')
        self.fake_server = fakes.make_fake_server(
            self.server_id, self.server_name
        )

    def test_server_set_metadata_with_exception(self):
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', self.server_name],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', 'detail'],
                        qs_elements=[f'name={self.server_name}'],
                    ),
                    json={'servers': [self.fake_server]},
                ),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', self.fake_server['id'], 'metadata'],
                    ),
                    validate=dict(json={'metadata': {'meta': 'data'}}),
                    json={},
                    status_code=400,
                ),
            ]
        )

        self.assertRaises(
            exceptions.BadRequestException,
            self.cloud.set_server_metadata,
            self.server_name,
            {'meta': 'data'},
        )

        self.assert_calls()

    def test_server_set_metadata(self):
        metadata = {'meta': 'data'}
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', self.server_id]
                    ),
                    json={'server': self.fake_server},
                ),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', self.fake_server['id'], 'metadata'],
                    ),
                    validate=dict(json={'metadata': metadata}),
                    status_code=200,
                    json={'metadata': metadata},
                ),
            ]
        )

        self.cloud.set_server_metadata(self.server_id, metadata)

        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import errno
import os

from examples.connect import FLAVOR_NAME
from examples.connect import IMAGE_NAME
from examples.connect import KEYPAIR_NAME
from examples.connect import NETWORK_NAME
from examples.connect import PRIVATE_KEYPAIR_FILE
from examples.connect import SERVER_NAME
from examples.connect import SSH_DIR

"""
Create resources with the Compute service.

For a full guide see
https://docs.openstack.org/openstacksdk/latest/user/guides/compute.html
"""


def create_keypair(conn):
    keypair = conn.compute.find_keypair(KEYPAIR_NAME)

    if not keypair:
        print("Create Key Pair:")

        keypair = conn.compute.create_keypair(name=KEYPAIR_NAME)

        print(keypair)

        try:
            os.mkdir(SSH_DIR)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise e

        with open(PRIVATE_KEYPAIR_FILE, 'w') as f:
            f.write(str(keypair.private_key))

        os.chmod(PRIVATE_KEYPAIR_FILE, 0o400)

    return keypair


def create_server(conn):
    print("Create Server:")

    image = conn.image.find_image(IMAGE_NAME)
    flavor = conn.compute.find_flavor(FLAVOR_NAME)
    network = conn.network.find_network(NETWORK_NAME)
    keypair = create_keypair(conn)

    server = conn.compute.create_server(
        name=SERVER_NAME,
        image_id=image.id,
        flavor_id=flavor.id,
        networks=[{"uuid": network.id}],
        key_name=keypair.name,
    )

    server = conn.compute.wait_for_server(server)

    print(f"ssh -i {PRIVATE_KEYPAIR_FILE} root@{server.access_ipv4}")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.shared_file_system.v2 import share_replica
from openstack.tests.unit import base

EXAMPLE = {
    "share_id": "9faf6f83-24c2-4551-9945-ca35f30a6ff7",
    "status": "available",
    "cast_rules_to_readonly": False,
    "updated_at": "2021-10-27T23:34:31.066083",
    "share_network_id": None,
    "share_server_id": None,
    "host": "some_random_host",
    "id": "6f406701-eb8c-400d-9ef0-1cda205a917d",
    "replica_state": "active",
    "created_at": "2021-10-27T23:34:26.350382",
}


class TestShareReplica(base.TestCase):
    def setUp(self):
        return super().setUp()

    def test_basic(self):
        sr = share_replica.ShareReplica()
        self.assertEqual('share_replica', sr.resource_key)
        self.assertEqual('share_replicas', sr.resources_key)
        self.assertEqual('share-replicas', sr.base_path)
        self.assertTrue(sr.allow_create)
        self.assertTrue(sr.allow_fetch)
        self.assertFalse(sr.allow_commit)
        self.assertTrue(sr.allow_delete)
        self.assertTrue(sr.allow_list)
        self.assertFalse(sr.allow_head)

    def test_share_replica(self):
        sr = share_replica.ShareReplica(**EXAMPLE)
        self.assertEqual(EXAMPLE['share_id'], sr.share_id)
        self.assertEqual(EXAMPLE['status'], sr.status)
        self.assertEqual(
            EXAMPLE['cast_rules_to_readonly'], sr.cast_rules_to_readonly
        )
        self.assertEqual(EXAMPLE['updated_at'], sr.updated_at)
        self.assertEqual(EXAMPLE['share_network_id'], sr.share_network_id)
        self.assertEqual(EXAMPLE['share_server_id'], sr.share_server_id)
        self.assertEqual(EXAMPLE['host'], sr.host)
        self.assertEqual(EXAMPLE['id'], sr.id)
        self.assertEqual(EXAMPLE['replica_state'], sr.replica_state)
        self.assertEqual(EXAMPLE['created_at'], sr.created_at)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from testtools import matchers

from openstack import exceptions
from openstack.tests.unit import base


class TestProject(base.TestCase):
    def get_mock_url(
        self,
        service_type='identity',
        interface='public',
        resource=None,
        append=None,
        base_url_append=None,
        v3=True,
        qs_elements=None,
    ):
        if v3 and resource is None:
            resource = 'projects'
        elif not v3 and resource is None:
            resource = 'tenants'
        if base_url_append is None and v3:
            base_url_append = 'v3'
        return super().get_mock_url(
            service_type=service_type,
            interface=interface,
            resource=resource,
            append=append,
            base_url_append=base_url_append,
            qs_elements=qs_elements,
        )

    def test_create_project_v3(
        self,
    ):
        project_data = self._get_project_data(
            description=self.getUniqueString('projectDesc'),
            parent_id=uuid.uuid4().hex,
        )
        reference_req = project_data.json_request.copy()
        reference_req['project']['enabled'] = True
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(),
                    status_code=200,
                    json=project_data.json_response,
                    validate=dict(json=reference_req),
                )
            ]
        )
        project = self.cloud.create_project(
            name=project_data.project_name,
            description=project_data.description,
            domain_id=project_data.domain_id,
            parent_id=project_data.parent_id,
        )
        self.assertThat(project.id, matchers.Equals(project_data.project_id))
        self.assertThat(
            project.name, matchers.Equals(project_data.project_name)
        )
        self.assertThat(
            project.description, matchers.Equals(project_data.description)
        )
        self.assertThat(
            project.domain_id, matchers.Equals(project_data.domain_id)
        )
        self.assert_calls()

    def test_delete_project_v3(self):
        project_data = self._get_project_data(v3=False)
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(append=[project_data.project_id]),
                    status_code=200,
                    json=project_data.json_response,
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(append=[project_data.project_id]),
                    status_code=204,
                ),
            ]
        )
        self.cloud.delete_project(project_data.project_id)
        self.assert_calls()

    def test_update_project_not_found(self):
        project_data = self._get_project_data()
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(append=[project_data.project_id]),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        qs_elements=['name=' + project_data.project_id]
                    ),
                    status_code=200,
                    json={'projects': []},
                ),
            ]
        )
        with self.assertRaises(exceptions.NotFoundException):
            self.cloud.update_project(project_data.project_id)
        self.assert_calls()

    def test_update_project_v3(self):
        project_data = self._get_project_data(
            description=self.getUniqueString('projectDesc')
        )
        reference_req = project_data.json_request.copy()
        # Remove elements not actually sent in the update
        reference_req['project'].pop('domain_id')
        reference_req['project'].pop('name')
        reference_req['project'].pop('enabled')
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        append=[project_data.project_id],
                        qs_elements=['domain_id=' + project_data.domain_id],
                    ),
                    status_code=200,
                    json={'projects': [project_data.json_response['project']]},
                ),
                dict(
                    method='PATCH',
                    uri=self.get_mock_url(append=[project_data.project_id]),
                    status_code=200,
                    json=project_data.json_response,
                    validate=dict(json=reference_req),
                ),
            ]
        )
        project = self.cloud.update_project(
            project_data.project_id,
            description=project_data.description,
            domain_id=project_data.domain_id,
        )
        self.assertThat(project.id, matchers.Equals(project_data.project_id))
        self.assertThat(
            project.name, matchers.Equals(project_data.project_name)
        )
        self.assertThat(
            project.description, matchers.Equals(project_data.description)
        )
        self.assert_calls()

    def test_list_projects_v3(self):
        project_data = self._get_project_data(
            description=self.getUniqueString('projectDesc')
        )
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource=(
                            f'projects?domain_id={project_data.domain_id}'
                        )
                    ),
                    status_code=200,
                    json={'projects': [project_data.json_response['project']]},
                )
            ]
        )
        projects = self.cloud.list_projects(project_data.domain_id)
        self.assertThat(len(projects), matchers.Equals(1))
        self.assertThat(
            projects[0].id, matchers.Equals(project_data.project_id)
        )
        self.assert_calls()

    def test_list_projects_v3_kwarg(self):
        project_data = self._get_project_data(
            description=self.getUniqueString('projectDesc')
        )
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource=(
                            f'projects?domain_id={project_data.domain_id}'
                        )
                    ),
                    status_code=200,
                    json={'projects': [project_data.json_response['project']]},
                )
            ]
        )
        projects = self.cloud.list_projects(domain_id=project_data.domain_id)
        self.assertThat(len(projects), matchers.Equals(1))
        self.assertThat(
            projects[0].id, matchers.Equals(project_data.project_id)
        )
        self.assert_calls()

    def test_list_projects_search_compat(self):
        project_data = self._get_project_data(
            description=self.getUniqueString('projectDesc')
        )
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(),
                    status_code=200,
                    json={'projects': [project_data.json_response['project']]},
                )
            ]
        )
        projects = self.cloud.search_projects(project_data.project_id)
        self.assertThat(len(projects), matchers.Equals(1))
        self.assertThat(
            projects[0].id, matchers.Equals(project_data.project_id)
        )
        self.assert_calls()

    def test_list_projects_search_compat_v3(self):
        project_data = self._get_project_data(
            description=self.getUniqueString('projectDesc')
        )
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource=(
                            f'projects?domain_id={project_data.domain_id}'
                        )
                    ),
                    status_code=200,
                    json={'projects': [project_data.json_response['project']]},
                )
            ]
        )
        projects = self.cloud.search_projects(domain_id=project_data.domain_id)
        self.assertThat(len(projects), matchers.Equals(1))
        self.assertThat(
            projects[0].id, matchers.Equals(project_data.project_id)
        )
        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import Any, cast

from keystoneauth1 import adapter

from openstack.clustering.v1 import _async_resource
from openstack import resource
from openstack import utils


class Node(_async_resource.AsyncResource):
    resource_key = 'node'
    resources_key = 'nodes'
    base_path = '/nodes'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    commit_method = 'PATCH'

    _query_mapping = resource.QueryParameters(
        'show_details',
        'name',
        'sort',
        'global_project',
        'cluster_id',
        'status',
    )

    # Properties
    #: The name of the node.
    name = resource.Body('name')
    #: The ID of the physical object that backs the node.
    physical_id = resource.Body('physical_id')
    #: The ID of the cluster in which this node is a member.
    #: A node is an orphan node if this field is empty.
    cluster_id = resource.Body('cluster_id')
    #: The ID of the profile used by this node.
    profile_id = resource.Body('profile_id')
    #: The domain ID of the node.
    domain_id = resource.Body('domain')
    #: The ID of the user who created this node.
    user_id = resource.Body('user')
    #: The ID of the project this node belongs to.
    project_id = resource.Body('project')
    #: The name of the profile used by this node.
    profile_name = resource.Body('profile_name')
    #: An integer that is unique inside the owning cluster.
    #: A value of -1 means this node is an orphan node.
    index = resource.Body('index', type=int)
    #: A string indicating the role the node plays in a cluster.
    role = resource.Body('role')
    #: The timestamp of the node object's initialization.
    #: *Type: datetime object parsed from ISO 8601 formatted string*
    init_at = resource.Body('init_at')
    #: The timestamp of the node's creation, i.e. the physical object
    #: represented by this node is also created.
    #: *Type: datetime object parsed from ISO 8601 formatted string*
    created_at = resource.Body('created_at')
    #: The timestamp the node was last updated.
    #: *Type: datetime object parsed from ISO 8601 formatted string*
    updated_at = resource.Body('updated_at')
    #: A string indicating the node's status.
    status = resource.Body('status')
    #: A string describing why the node entered its current status.
    status_reason = resource.Body('status_reason')
    #: A map containing key-value pairs attached to the node.
    metadata = resource.Body('metadata', type=dict)
    #: A map containing some runtime data for this node.
    data = resource.Body('data', type=dict)
    #: A map containing the details of the physical object this node
    #: represents
    details = resource.Body('details', type=dict)
    #: A map containing the dependency of nodes
    dependents = resource.Body('dependents', type=dict)
    #: Whether the node is tainted. *Type: bool*
    tainted = resource.Body('tainted', type=bool)

    def _action(
        self, session: adapter.Adapter, body: dict[str, Any]
    ) -> dict[str, Any]:
        """Procedure the invoke an action API.

        :param session: A session object used for sending request.
        :param body: The body of action to be sent.
        """
        url = utils.urljoin(self.base_path, self.id, 'actions')
        resp = session.post(url, json=body)
        return cast(dict[str, Any], resp.json())

    def check(self, session: adapter.Adapter, **params: Any) -> dict[str, Any]:
        """An action procedure for the node to check its health status.

        :param session: A session object used for sending request.
        :returns: A dictionary containing the action ID.
        """
        body = {'check': params}
        return self._action(session, body)

    def recover(
        self, session: adapter.Adapter, **params: Any
    ) -> dict[str, Any]:
        """An action procedure for the node to recover.

        :param session: A session object used for sending request.
        :returns: A dictionary containing the action ID.
        """
        body = {'recover': params}
        return self._action(session, body)

    def op(
        self, session: adapter.Adapter, operation: str, **params: Any
    ) -> dict[str, Any]:
        """Perform an operation on the specified node.

        :param session: A session object used for sending request.
        :param operation: A string representing the operation to be performed.
        :param dict params: An optional dict providing the parameters for the
                            operation.
        :returns: A dictionary containing the action ID.
        """
        url = utils.urljoin(self.base_path, self.id, 'ops')
        resp = session.post(url, json={operation: params})
        return cast(dict[str, Any], resp.json())

    def adopt(
        self,
        session: adapter.Adapter,
        preview: bool = False,
        **params: Any,
    ) -> dict[str, Any] | Node:
        """Adopt a node for management.

        :param session: A session object used for sending request.
        :param preview: A boolean indicating whether the adoption is a
                        preview. A "preview" does not create the node object.
        :param dict params: A dict providing the details of a node to be
                            adopted.
        """
        if preview:
            path = 'adopt-preview'
            attrs: dict[str, Any] = {
                'identity': params.get('identity'),
                'overrides': params.get('overrides'),
                'type': params.get('type'),
                'snapshot': params.get('snapshot'),
            }
        else:
            path = 'adopt'
            attrs = params

        url = utils.urljoin(self.base_path, path)
        resp = session.post(url, json=attrs)
        if preview:
            return cast(dict[str, Any], resp.json())

        self._translate_response(resp)
        return self


class NodeDetail(Node):
    base_path = '/nodes/%(node_id)s?show_details=True'

    allow_create = False
    allow_fetch = True
    allow_commit = False
    allow_delete = False
    allow_list = False

    node_id = resource.URI('node_id')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.tests.functional import base


class BaseBaremetalTest(base.BaseFunctionalTest):
    min_microversion: str | None = None
    node_id: str

    def setUp(self):
        super().setUp()
        self.require_service(
            'baremetal', min_microversion=self.min_microversion
        )

    def create_allocation(self, **kwargs):
        allocation = self.operator_cloud.baremetal.create_allocation(**kwargs)
        self.addCleanup(
            lambda: self.operator_cloud.baremetal.delete_allocation(
                allocation.id, ignore_missing=True
            )
        )
        return allocation

    def create_chassis(self, **kwargs):
        chassis = self.system_admin_cloud.baremetal.create_chassis(**kwargs)
        self.addCleanup(
            lambda: self.system_admin_cloud.baremetal.delete_chassis(
                chassis.id, ignore_missing=True
            )
        )
        return chassis

    def create_node(self, driver='fake-hardware', **kwargs):
        node = self.operator_cloud.baremetal.create_node(
            driver=driver, **kwargs
        )
        self.node_id = node.id
        self.addCleanup(
            lambda: self.operator_cloud.baremetal.delete_node(
                self.node_id, ignore_missing=True
            )
        )
        self.assertIsNotNone(self.node_id)
        return node

    def create_port(self, node_id=None, **kwargs):
        node_id = node_id or self.node_id
        port = self.operator_cloud.baremetal.create_port(
            node_uuid=node_id, **kwargs
        )
        self.addCleanup(
            lambda: self.operator_cloud.baremetal.delete_port(
                port.id, ignore_missing=True
            )
        )
        return port

    def create_port_group(self, node_id=None, **kwargs):
        node_id = node_id or self.node_id
        port_group = self.operator_cloud.baremetal.create_port_group(
            node_uuid=node_id, **kwargs
        )
        self.addCleanup(
            lambda: self.operator_cloud.baremetal.delete_port_group(
                port_group.id, ignore_missing=True
            )
        )
        return port_group

    def create_volume_connector(self, node_id=None, **kwargs):
        node_id = node_id or self.node_id
        volume_connector = (
            self.operator_cloud.baremetal.create_volume_connector(
                node_uuid=node_id, **kwargs
            )
        )

        self.addCleanup(
            lambda: self.operator_cloud.baremetal.delete_volume_connector(
                volume_connector.id, ignore_missing=True
            )
        )
        return volume_connector

    def create_volume_target(self, node_id=None, **kwargs):
        node_id = node_id or self.node_id
        volume_target = self.operator_cloud.baremetal.create_volume_target(
            node_uuid=node_id, **kwargs
        )

        self.addCleanup(
            lambda: self.operator_cloud.baremetal.delete_volume_target(
                volume_target.id, ignore_missing=True
            )
        )
        return volume_target

    def create_deploy_template(self, **kwargs):
        """Create a new deploy_template from attributes."""

        deploy_template = (
            self.system_admin_cloud.baremetal.create_deploy_template(**kwargs)
        )

        self.addCleanup(
            lambda: self.system_admin_cloud.baremetal.delete_deploy_template(
                deploy_template.id, ignore_missing=True
            )
        )
        return deploy_template

    def create_runbook(self, **kwargs):
        """Create a new runbook from attributes."""

        runbook = self.operator_cloud.baremetal.create_runbook(**kwargs)

        self.addCleanup(
            lambda: self.operator_cloud.baremetal.delete_runbook(
                runbook.id, ignore_missing=True
            )
        )
        return runbook

    def create_inspection_rule(self, **kwargs):
        """Create a new inspection_rule from attributes."""

        inspection_rule = (
            self.system_admin_cloud.baremetal.create_inspection_rule(**kwargs)
        )

        self.addCleanup(
            lambda: self.system_admin_cloud.baremetal.delete_inspection_rule(
                inspection_rule.id, ignore_missing=True
            )
        )
        return inspection_rule



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Helpers for building configdrive compatible with the Bare Metal service."""

import base64
import contextlib
from collections.abc import Generator, Iterable, Mapping
import gzip
import json
import os
import shutil
import subprocess
import tempfile


@contextlib.contextmanager
def populate_directory(
    metadata: Mapping[str, object],
    user_data: bytes | None = None,
    versions: Iterable[str] | None = None,
    network_data: Mapping[str, object] | None = None,
    vendor_data: Mapping[str, object] | None = None,
) -> Generator[str, None, None]:
    """Populate a directory with configdrive files.

    :param metadata: Metadata.
    :param user_data: Vendor-specific user data.
    :param versions: List of metadata versions to support.
    :param network_data: Networking configuration.
    :param vendor_data: Extra supplied vendor data.
    :return: a context manager yielding a directory with files
    """
    d = tempfile.mkdtemp()
    versions = versions or ('2012-08-10', 'latest')
    try:
        for version in versions:
            subdir = os.path.join(d, 'openstack', version)
            if not os.path.exists(subdir):
                os.makedirs(subdir)

            with open(os.path.join(subdir, 'meta_data.json'), 'w') as fp:
                json.dump(metadata, fp)

            if network_data:
                with open(
                    os.path.join(subdir, 'network_data.json'), 'w'
                ) as fp:
                    json.dump(network_data, fp)

            if vendor_data:
                with open(
                    os.path.join(subdir, 'vendor_data2.json'), 'w'
                ) as fp:
                    json.dump(vendor_data, fp)

            if user_data:
                # Strictly speaking, user data is binary, but in many cases
                # it's actually a text (cloud-init, ignition, etc).
                flag = 't' if isinstance(user_data, str) else 'b'
                with open(os.path.join(subdir, 'user_data'), f'w{flag}') as fp:
                    fp.write(user_data)

        yield d
    finally:
        shutil.rmtree(d)


def build(
    metadata: Mapping[str, object],
    user_data: bytes | None = None,
    versions: Iterable[str] | None = None,
    network_data: Mapping[str, object] | None = None,
    vendor_data: Mapping[str, object] | None = None,
) -> str:
    """Make a configdrive compatible with the Bare Metal service.

    Requires the genisoimage utility to be available.

    :param metadata: Metadata.
    :param user_data: Vendor-specific user data.
    :param versions: List of metadata versions to support.
    :param network_data: Networking configuration.
    :param vendor_data: Extra supplied vendor data.
    :return: configdrive contents as a base64-encoded string.
    """
    with populate_directory(
        metadata, user_data, versions, network_data, vendor_data
    ) as path:
        return pack(path)


def pack(path: str) -> str:
    """Pack a directory with files into a Bare Metal service configdrive.

    Creates an ISO image with the files and label "config-2".

    :param path: Path to directory with files
    :return: configdrive contents as a base64-encoded string.
    """
    with tempfile.NamedTemporaryFile() as tmpfile:
        # NOTE(toabctl): Luckily, genisoimage, mkisofs and xorrisofs understand
        # the same parameters which are currently used.
        error: Exception | None
        for c in ['genisoimage', 'mkisofs', 'xorrisofs']:
            try:
                p = subprocess.Popen(  # noqa: S603
                    [
                        c,
                        '-o',
                        tmpfile.name,
                        '-ldots',
                        '-allow-lowercase',
                        '-allow-multidot',
                        '-l',
                        '-publisher',
                        'metalsmith',
                        '-quiet',
                        '-J',
                        '-r',
                        '-V',
                        'config-2',
                        path,
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
            except OSError as e:
                error = e
            else:
                error = None
                break

        if error:
            raise RuntimeError(
                'Error generating the configdrive. Make sure the '
                '"genisoimage", "mkisofs" or "xorrisofs" tool is installed. '
                f'Error: {error}'
            )

        stdout, stderr = p.communicate()
        if p.returncode != 0:
            raise RuntimeError(
                'Error generating the configdrive.'
                f'Stdout: "{stdout.decode()}". Stderr: "{stderr.decode()}"'
            )

        tmpfile.seek(0)

        with tempfile.NamedTemporaryFile() as tmpzipfile:
            with gzip.GzipFile(fileobj=tmpzipfile, mode='wb') as gz_file:
                shutil.copyfileobj(tmpfile, gz_file)

            tmpzipfile.seek(0)
            # NOTE(dtantsur): Ironic expects configdrive to be a string, but
            # base64 returns bytes on Python 3.
            cd = base64.b64encode(tmpzipfile.read()).decode()

    return cd



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional.shared_file_system.v2 import base


class TestExportLocation(base.BaseSharedFileSystemTest):
    min_microversion = '2.9'

    def setUp(self):
        super().setUp()

        self.SHARE_NAME = self.getUniqueString()
        my_share = self.create_share(
            name=self.SHARE_NAME,
            size=2,
            share_type="dhss_false",
            share_protocol='NFS',
            description=None,
        )
        self.SHARE_ID = my_share.id

    def test_export_locations(self):
        exs = list(
            self.user_cloud.shared_file_system.export_locations(self.SHARE_ID)
        )
        self.assertGreater(len(exs), 0)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.image.v1 import _proxy
from openstack.image.v1 import image
from openstack.tests.unit import test_proxy_base as test_proxy_base


class TestImageProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_image_upload_attrs(self):
        self.verify_create(self.proxy.upload_image, image.Image)

    def test_image_delete(self):
        self.verify_delete(self.proxy.delete_image, image.Image, False)

    def test_image_delete_ignore(self):
        self.verify_delete(self.proxy.delete_image, image.Image, True)

    def test_image_find(self):
        self.verify_find(self.proxy.find_image, image.Image)

    def test_image_get(self):
        self.verify_get(self.proxy.get_image, image.Image)

    def test_images(self):
        self.verify_list(self.proxy.images, image.Image)

    def test_image_update(self):
        self.verify_update(self.proxy.update_image, image.Image)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional import base


class TestAutoAllocatedTopology(base.BaseFunctionalTest):
    NETWORK_NAME = "auto_allocated_network"
    NETWORK_ID: str
    PROJECT_ID: str

    def setUp(self):
        super().setUp()

        if not self.operator_cloud:
            self.skipTest("Operator cloud is required for this test")

        if not self.operator_cloud._has_neutron_extension(
            "auto-allocated-topology"
        ):
            self.skipTest(
                "Neutron auto-allocated-topology extension is "
                "required for this test"
            )

        project = self._create_project()
        self.PROJECT_ID = project['id']
        self.test_cloud = self.operator_cloud.connect_as_project(project)

        # Dry run will only pass if there is a public network
        self._set_network_external()

    def tearDown(self):
        res = self.test_cloud.network.delete_auto_allocated_topology(
            self.PROJECT_ID
        )
        self.assertIsNone(res)
        self._destroy_project()
        super().tearDown()

    def _create_project(self):
        project_name = 'auto_allocated_topology_test_project'
        project = self.operator_cloud.get_project(project_name)
        if not project:
            params = {
                'name': project_name,
                'description': (
                    'test project used only for the '
                    'TestAutoAllocatedTopology tests class'
                ),
                'domain_id': self.operator_cloud.get_domain('default')['id'],
            }

            project = self.operator_cloud.create_project(**params)

        user_id = self.operator_cloud.current_user_id
        # Grant the current user access to the project
        role_assignment = self.operator_cloud.list_role_assignments(
            {'user': user_id, 'project': project['id']}
        )
        if not role_assignment:
            self.operator_cloud.grant_role(
                'member', user=user_id, project=project['id'], wait=True
            )
        return project

    def _destroy_project(self):
        self.operator_cloud.revoke_role(
            'member',
            user=self.operator_cloud.current_user_id,
            project=self.PROJECT_ID,
        )
        self.operator_cloud.delete_project(self.PROJECT_ID)

    def test_auto_allocated_topology(self):
        # First test validation with the 'dry-run' call
        # Dry run option will return "dry-run=pass" in the 'id' resource
        sop = self.test_cloud.network.validate_auto_allocated_topology(
            self.PROJECT_ID
        )
        self.assertEqual(self.PROJECT_ID, sop.project)
        self.assertEqual("dry-run=pass", sop.id)

        # test show auto_allocated_network without project id in the request
        top = self.test_cloud.network.get_auto_allocated_topology()
        project = self.test_cloud.session.get_project_id()
        network = self.test_cloud.network.get_network(top.id)
        self.assertEqual(top.project_id, project)
        self.assertEqual(top.id, network.id)

        # test show auto_allocated_network with project id in the request
        top = self.test_cloud.network.get_auto_allocated_topology(
            self.PROJECT_ID
        )
        network = self.test_cloud.network.get_network(top.id)
        self.assertEqual(top.project_id, network.project_id)
        self.assertEqual(top.id, network.id)
        self.assertEqual(network.name, "auto_allocated_network")

    def _set_network_external(self):
        networks = self.test_cloud.network.networks()
        for network in networks:
            if network.name == "public":
                self.test_cloud.network.update_network(
                    network, is_default=True
                )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Self

from keystoneauth1 import adapter

from openstack import resource


class Deployable(resource.Resource):
    resource_key = 'deployable'
    resources_key = 'deployables'
    base_path = '/deployables'
    # capabilities
    allow_create = False
    allow_fetch = True
    allow_commit = False
    allow_delete = False
    allow_list = True
    allow_patch = True
    #: The timestamp when this deployable was created.
    created_at = resource.Body('created_at')
    #: The device_id of the deployable.
    device_id = resource.Body('device_id')
    #: The UUID of the deployable.
    id = resource.Body('uuid', alternate_id=True)
    #: The name of the deployable.
    name = resource.Body('name')
    #: The num_accelerator of the deployable.
    num_accelerators = resource.Body('num_accelerators')
    #: The parent_id of the deployable.
    parent_id = resource.Body('parent_id')
    #: The root_id of the deployable.
    root_id = resource.Body('root_id')
    #: The timestamp when this deployable was updated.
    updated_at = resource.Body('updated_at')

    def patch(
        self,
        session: adapter.Adapter,
        patch: list[dict[str, Any]] | None = None,
        prepend_key: bool = True,
        has_body: bool = True,
        retry_on_conflict: bool | None = None,
        base_path: str | None = '/deployables/%(id)s/program',
        *,
        microversion: str | None = None,
    ) -> Self:
        return super().patch(
            session,
            patch=patch,
            prepend_key=prepend_key,
            has_body=has_body,
            retry_on_conflict=retry_on_conflict,
            base_path=base_path,
            microversion=microversion,
        )



# Copyright (c) 2018 China Telecom Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from openstack.load_balancer.v2 import quota
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'load_balancer': 1,
    'listener': 2,
    'pool': 3,
    'health_monitor': 4,
    'member': 5,
    'project_id': 6,
}


class TestQuota(base.TestCase):
    def test_basic(self):
        sot = quota.Quota()
        self.assertEqual('quota', sot.resource_key)
        self.assertEqual('quotas', sot.resources_key)
        self.assertEqual('/lbaas/quotas', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = quota.Quota(**EXAMPLE)
        self.assertEqual(EXAMPLE['load_balancer'], sot.load_balancers)
        self.assertEqual(EXAMPLE['listener'], sot.listeners)
        self.assertEqual(EXAMPLE['pool'], sot.pools)
        self.assertEqual(EXAMPLE['health_monitor'], sot.health_monitors)
        self.assertEqual(EXAMPLE['member'], sot.members)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)

    def test_prepare_request(self):
        body = {'id': 'ABCDEFGH', 'load_balancer': '12345'}
        quota_obj = quota.Quota(**body)
        response = quota_obj._prepare_request()
        self.assertNotIn('id', response)


class TestQuotaDefault(base.TestCase):
    def test_basic(self):
        sot = quota.QuotaDefault()
        self.assertEqual('quota', sot.resource_key)
        self.assertEqual('quotas', sot.resources_key)
        self.assertEqual('/lbaas/quotas/defaults', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_list)
        self.assertTrue(sot.allow_retrieve)

    def test_make_it(self):
        sot = quota.Quota(**EXAMPLE)
        self.assertEqual(EXAMPLE['load_balancer'], sot.load_balancers)
        self.assertEqual(EXAMPLE['listener'], sot.listeners)
        self.assertEqual(EXAMPLE['pool'], sot.pools)
        self.assertEqual(EXAMPLE['health_monitor'], sot.health_monitors)
        self.assertEqual(EXAMPLE['member'], sot.members)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class ProjectQuota(resource.Resource):
    resource_key = 'project_quotas'
    resources_key = 'project_quotas'
    base_path = '/project-quotas'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    # Properties
    #: Contains the configured quota value of the requested project for the
    #: secret resource.
    secrets = resource.Body("secrets")
    #: Contains the configured quota value of the requested project for the
    #: orders resource.
    orders = resource.Body("orders")
    #: Contains the configured quota value of the requested project for the
    #: containers resource.
    containers = resource.Body("containers")
    #: Contains the configured quota value of the requested project for the
    #: consumers resource.
    consumers = resource.Body("consumers")
    #: Contains the configured quota value of the requested project for the CAs
    #: resource.
    cas = resource.Body("cas")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Callable, Generator
from typing import Any, ClassVar, Literal, overload
import warnings

from openstack.placement.v1 import resource_class as _resource_class
from openstack.placement.v1 import resource_provider as _resource_provider
from openstack.placement.v1 import (
    resource_provider_inventory as _resource_provider_inventory,
)
from openstack.placement.v1 import trait as _trait
from openstack import proxy
from openstack import resource
from openstack import warnings as os_warnings


class Proxy(proxy.Proxy):
    api_version: ClassVar[Literal['1']] = '1'

    _resource_registry = {
        "resource_class": _resource_class.ResourceClass,
        "resource_provider": _resource_provider.ResourceProvider,
    }

    # resource classes

    def create_resource_class(
        self, **attrs: Any
    ) -> _resource_class.ResourceClass:
        """Create a new resource class from attributes.

        :param attrs: Keyword arguments which will be used to create a
            :class:`~openstack.placement.v1.resource_provider.ResourceClass`,
            comprised of the properties on the ResourceClass class.

        :returns: The results of resource class creation
        """
        return self._create(_resource_class.ResourceClass, **attrs)

    def delete_resource_class(
        self,
        resource_class: str | _resource_class.ResourceClass,
        ignore_missing: bool = True,
    ) -> None:
        """Delete a resource class

        :param resource_class: The value can be either the ID of a resource
            class or an
            :class:`~openstack.placement.v1.resource_class.ResourceClass`,
            instance.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the resource class does not exist. When set to ``True``, no
            exception will be set when attempting to delete a nonexistent
            resource class.

        :returns: ``None``
        """
        self._delete(
            _resource_class.ResourceClass,
            resource_class,
            ignore_missing=ignore_missing,
        )

    def update_resource_class(
        self, resource_class: str | _resource_class.ResourceClass, **attrs: Any
    ) -> _resource_class.ResourceClass:
        """Update a resource class

        :param resource_class: The value can be either the ID of a resource
            class or an
            :class:`~openstack.placement.v1.resource_class.ResourceClass`,
            instance.
        :param attrs: The attributes to update on the resource class
            represented by ``resource_class``.

        :returns: The updated resource class
        """
        return self._update(
            _resource_class.ResourceClass,
            resource_class,
            **attrs,
        )

    def get_resource_class(
        self,
        resource_class: str | _resource_class.ResourceClass,
    ) -> _resource_class.ResourceClass:
        """Get a single resource_class.

        :param resource_class: The value can be either the ID of a resource
            class or an
            :class:`~openstack.placement.v1.resource_class.ResourceClass`,
            instance.

        :returns: An instance of
            :class:`~openstack.placement.v1.resource_class.ResourceClass`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource class matching the criteria could be found.
        """
        return self._get(
            _resource_class.ResourceClass,
            resource_class,
        )

    def resource_classes(
        self,
        **query: Any,
    ) -> Generator[_resource_class.ResourceClass, None, None]:
        """Retrieve a generator of resource classs.

        :param query: Optional query parameters to be sent to
            restrict the resource classs to be returned.

        :returns: A generator of resource class instances.
        """
        return self._list(_resource_class.ResourceClass, **query)

    # resource providers

    def create_resource_provider(
        self, **attrs: Any
    ) -> _resource_provider.ResourceProvider:
        """Create a new resource provider from attributes.

        :param attrs: Keyword arguments which will be used to create a
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`,
            comprised of the properties on the ResourceProvider class.

        :returns: The results of resource provider creation
        """
        return self._create(_resource_provider.ResourceProvider, **attrs)

    def delete_resource_provider(
        self,
        resource_provider: str | _resource_provider.ResourceProvider,
        ignore_missing: bool = True,
    ) -> None:
        """Delete a resource provider

        :param resource_provider: The value can be either the ID of a resource
            provider or an
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`,
            instance.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the resource provider does not exist. When set to ``True``, no
            exception will be set when attempting to delete a nonexistent
            resource provider.

        :returns: ``None``
        """
        self._delete(
            _resource_provider.ResourceProvider,
            resource_provider,
            ignore_missing=ignore_missing,
        )

    def update_resource_provider(
        self,
        resource_provider: str | _resource_provider.ResourceProvider,
        **attrs: Any,
    ) -> _resource_provider.ResourceProvider:
        """Update a resource provider

        :param resource_provider: The value can be either the ID of a resource
            provider or an
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`,
            instance.
        :param attrs: The attributes to update on the resource provider
            represented by ``resource_provider``.

        :returns: The updated resource provider
        """
        return self._update(
            _resource_provider.ResourceProvider,
            resource_provider,
            **attrs,
        )

    def get_resource_provider(
        self,
        resource_provider: str | _resource_provider.ResourceProvider,
    ) -> _resource_provider.ResourceProvider:
        """Get a single resource_provider.

        :param resource_provider: The value can be either the ID of a resource
            provider or an
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`,
            instance.

        :returns: An instance of
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource provider matching the criteria could be found.
        """
        return self._get(
            _resource_provider.ResourceProvider,
            resource_provider,
        )

    @overload
    def find_resource_provider(
        self,
        name_or_id: str,
        ignore_missing: Literal[False],
    ) -> _resource_provider.ResourceProvider: ...

    @overload
    def find_resource_provider(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _resource_provider.ResourceProvider | None: ...

    def find_resource_provider(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _resource_provider.ResourceProvider | None:
        """Find a single resource_provider.

        :param name_or_id: The name or ID of a resource provider.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the resource does not exist.  When set to ``True``, None will
            be returned when attempting to find a nonexistent resource.

        :returns: An instance of
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource provider matching the criteria could be found.
        """
        return self._find(
            _resource_provider.ResourceProvider,
            name_or_id,
            ignore_missing=ignore_missing,
        )

    def resource_providers(
        self,
        **query: Any,
    ) -> Generator[_resource_provider.ResourceProvider, None, None]:
        """Retrieve a generator of resource providers.

        :param query: Optional query parameters to be sent to
            restrict the resource providers to be returned.

        :returns: A generator of resource provider instances.
        """
        return self._list(_resource_provider.ResourceProvider, **query)

    # resource provider aggregates

    def fetch_resource_provider_aggregates(
        self,
        resource_provider: str | _resource_provider.ResourceProvider,
    ) -> _resource_provider.ResourceProvider:
        """Get a list of aggregates for a resource provider.

        :param resource_provider: The value can be either the ID of a resource
            provider or an
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`,
            instance.

        :returns: An instance of
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`
            with the ``aggregates`` attribute populated.
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource provider matching the criteria could be found.
        """
        res = self._get_resource(
            _resource_provider.ResourceProvider,
            resource_provider,
        )
        return res.fetch_aggregates(self)

    # TODO(stephenfin): Remove in 5.0
    def get_resource_provider_aggregates(
        self,
        resource_provider: str | _resource_provider.ResourceProvider,
    ) -> _resource_provider.ResourceProvider:
        """Get a list of aggregates for a resource provider.

        .. deprecated:: 4.14.0
            Use :meth:`fetch_resource_provider_aggregates` instead.
        """
        warnings.warn(
            "The 'get_resource_provider_aggregates' method is deprecated; "
            "use 'fetch_resource_provider_aggregates' instead.",
            os_warnings.RemovedInSDK50Warning,
        )
        return self.fetch_resource_provider_aggregates(resource_provider)

    def set_resource_provider_aggregates(
        self,
        resource_provider: str | _resource_provider.ResourceProvider,
        *aggregates: str,
    ) -> _resource_provider.ResourceProvider:
        """Update aggregates for a resource provider.

        :param resource_provider: The value can be either the ID of a resource
            provider or an
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`,
            instance.
        :param aggregates: A list of aggregates. These aggregates will replace
            all aggregates currently present.

        :returns: An instance of
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`
            with the ``aggregates`` attribute populated with the updated value.
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource provider matching the criteria could be found.
        """
        res = self._get_resource(
            _resource_provider.ResourceProvider,
            resource_provider,
        )
        return res.set_aggregates(self, aggregates=list(aggregates))

    # resource provider inventories

    def create_resource_provider_inventory(
        self,
        resource_provider: str | _resource_provider.ResourceProvider,
        resource_class: str | _resource_class.ResourceClass,
        *,
        total: int,
        **attrs: Any,
    ) -> _resource_provider_inventory.ResourceProviderInventory:
        """Create a new resource provider inventory from attributes

        :param resource_provider: Either the ID of a resource provider or a
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`
            instance.
        :param resource_class: The value can be either the ID of a resource
            class or an
            :class:`~openstack.placement.v1.resource_class.ResourceClass`,
            instance.
        :param total: The actual amount of the resource that the provider can
            accommodate.
        :param attrs: Keyword arguments which will be used to create a
            :class:`~openstack.placement.v1.resource_provider_inventory.ResourceProviderInventory`,
            comprised of the properties on the ResourceProviderInventory class.

        :returns: The results of resource provider inventory creation
        """
        resource_provider_id = resource.Resource._get_id(resource_provider)
        resource_class_name = resource.Resource._get_id(resource_class)
        return self._create(
            _resource_provider_inventory.ResourceProviderInventory,
            resource_provider_id=resource_provider_id,
            resource_class=resource_class_name,
            total=total,
            **attrs,
        )

    def delete_resource_provider_inventory(
        self,
        resource_provider_inventory: str
        | _resource_provider_inventory.ResourceProviderInventory,
        resource_provider: str
        | _resource_provider.ResourceProvider
        | None = None,
        ignore_missing: bool = True,
    ) -> None:
        """Delete a resource provider inventory

        :param resource_provider_inventory: The value can be either the ID of a
            resource provider or an
            :class:`~openstack.placement.v1.resource_provider_inventory.ResourceProviderInventory`,
            instance.
        :param resource_provider: Either the ID of a resource provider or a
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`
            instance. This value must be specified when
            ``resource_provider_inventory`` is an ID.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the resource provider inventory does not exist. When set to
            ``True``, no exception will be set when attempting to delete a
            nonexistent resource provider inventory.

        :returns: ``None``
        """
        resource_provider_id = self._get_uri_attribute(
            resource_provider_inventory,
            resource_provider,
            'resource_provider_id',
        )
        self._delete(
            _resource_provider_inventory.ResourceProviderInventory,
            resource_provider_inventory,
            resource_provider_id=resource_provider_id,
            ignore_missing=ignore_missing,
        )

    def update_resource_provider_inventory(
        self,
        resource_provider_inventory: str
        | _resource_provider_inventory.ResourceProviderInventory,
        resource_provider: str
        | _resource_provider.ResourceProvider
        | None = None,
        *,
        resource_provider_generation: int | None = None,
        **attrs: Any,
    ) -> _resource_provider_inventory.ResourceProviderInventory:
        """Update a resource provider's inventory

        :param resource_provider_inventory: The value can be either the ID of a resource
            provider inventory or an
            :class:`~openstack.placement.v1.resource_provider_inventory.ResourceProviderInventory`,
            instance.
        :param resource_provider: Either the ID of a resource provider or a
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`
            instance. This value must be specified when
            ``resource_provider_inventory`` is an ID.
        :attrs kwargs: The attributes to update on the resource provider inventory
            represented by ``resource_provider_inventory``.

        :returns: The updated resource provider inventory
        """  # noqa: E501
        resource_provider_id = self._get_uri_attribute(
            resource_provider_inventory,
            resource_provider,
            'resource_provider_id',
        )
        return self._update(
            _resource_provider_inventory.ResourceProviderInventory,
            resource_provider_inventory,
            resource_provider_id=resource_provider_id,
            resource_provider_generation=resource_provider_generation,
            **attrs,
        )

    def get_resource_provider_inventory(
        self,
        resource_provider_inventory: (
            str | _resource_provider_inventory.ResourceProviderInventory
        ),
        resource_provider: (
            str | _resource_provider.ResourceProvider | None
        ) = None,
    ) -> _resource_provider_inventory.ResourceProviderInventory:
        """Get a single resource_provider_inventory

        :param resource_provider_inventory: The value can be either the ID of a
            resource provider inventory or an
            :class:`~openstack.placement.v1.resource_provider_inventory.ResourceProviderInventory`,
            instance.
        :param resource_provider: Either the ID of a resource provider or a
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`
            instance. This value must be specified when
            ``resource_provider_inventory`` is an ID.

        :returns: An instance of
            :class:`~openstack.placement.v1.resource_provider_inventory.ResourceProviderInventory`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource provider inventory matching the criteria could be found.
        """
        resource_provider_id = self._get_uri_attribute(
            resource_provider_inventory,
            resource_provider,
            'resource_provider_id',
        )
        return self._get(
            _resource_provider_inventory.ResourceProviderInventory,
            resource_provider_inventory,
            resource_provider_id=resource_provider_id,
        )

    def resource_provider_inventories(
        self,
        resource_provider: str | _resource_provider.ResourceProvider,
        **query: Any,
    ) -> Generator[
        _resource_provider_inventory.ResourceProviderInventory, None, None
    ]:
        """Retrieve a generator of resource provider inventories

        :param resource_provider: Either the ID of a resource provider or a
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`
            instance.
        :param query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of resource provider inventory instances.
        """
        resource_provider_id = resource.Resource._get_id(resource_provider)
        return self._list(
            _resource_provider_inventory.ResourceProviderInventory,
            resource_provider_id=resource_provider_id,
            **query,
        )

    def fetch_resource_provider_usages(
        self,
        resource_provider: str | _resource_provider.ResourceProvider,
    ) -> _resource_provider.ResourceProvider:
        """Fetch usages for a specific resource provider.

        :param resource_provider: The value can be either the ID of a
            resource provider or a
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`
            instance.
        :returns: An instance of
            :class:`~openstack.placement.v1.resource_provider.ResourceProvider`
            with the ``usages`` attribute populated.
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource provider matching the criteria could be found.
        """
        res = self._get_resource(
            _resource_provider.ResourceProvider,
            resource_provider,
        )
        return res.fetch_usages(self)

    # ========== Traits ==========

    def create_trait(self, name: str) -> _trait.Trait:
        """Create a new trait

        :param name: The name of the new trait

        :returns: The results of trait creation
        """
        return self._create(_trait.Trait, name=name)

    def delete_trait(
        self, trait: str | _trait.Trait, ignore_missing: bool = True
    ) -> None:
        """Delete a trait

        :param trait: The value can be either the ID of a trait or an
            :class:`~openstack.placement.v1.trait.Trait`, instance.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the resource provider inventory does not exist. When set to
            ``True``, no exception will be set when attempting to delete a
            nonexistent resource provider inventory.

        :returns: ``None``
        """
        self._delete(_trait.Trait, trait, ignore_missing=ignore_missing)

    def get_trait(self, trait: str | _trait.Trait) -> _trait.Trait:
        """Get a single trait

        :param trait: The value can be either the ID of a trait or an
            :class:`~openstack.placement.v1.trait.Trait`, instance.

        :returns: An instance of
            :class:`~openstack.placement.v1.resource_provider_inventory.ResourceProviderInventory`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            trait matching the criteria could be found.
        """
        return self._get(_trait.Trait, trait)

    def traits(self, **query: Any) -> Generator[_trait.Trait, None, None]:
        """Retrieve a generator of traits

        :param query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of trait objects
        """
        return self._list(_trait.Trait, **query)

    # ========== Utilities ==========

    def wait_for_status(
        self,
        res: resource.ResourceT,
        status: str,
        failures: list[str] | None = None,
        interval: int | float | None = 2,
        wait: int | None = None,
        attribute: str = 'status',
        callback: Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for the resource to be in a particular status.

        :param session: The session to use for making this request.
        :param resource: The resource to wait on to reach the status. The
            resource must have a status attribute specified via ``attribute``.
        :param status: Desired status of the resource.
        :param failures: Statuses that would indicate the transition
            failed such as 'ERROR'. Defaults to ['ERROR'].
        :param interval: Number of seconds to wait between checks.
        :param wait: Maximum number of seconds to wait for transition.
            Set to ``None`` to wait forever.
        :param attribute: Name of the resource attribute that contains the
            status.
        :param callback: A callback function. This will be called with a single
            value, progress. This is API specific but is generally a percentage
            value from 0-100.

        :returns: The updated resource.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if the
            transition to status failed to occur in ``wait`` seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
            transitioned to one of the states in ``failures``.
        :raises: :class:`~AttributeError` if the resource does not have a
            ``status`` attribute
        """
        return resource.wait_for_status(
            self, res, status, failures, interval, wait, attribute, callback
        )

    def wait_for_delete(
        self,
        res: resource.ResourceT,
        interval: int | float | None = 2,
        wait: int | None = 120,
        callback: Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for a resource to be deleted.

        :param res: The resource to wait on to be deleted.
        :param interval: Number of seconds to wait before to consecutive
            checks.
        :param wait: Maximum number of seconds to wait before the change.
        :param callback: A callback function. This will be called with a single
            value, progress, which is a percentage value from 0-100.

        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
            to delete failed to occur in the specified seconds.
        """
        return resource.wait_for_delete(self, res, interval, wait, callback)



# Copyright (c) 2019, Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import re

from hacking import core

"""
Guidelines for writing new hacking checks

 - Use only for openstacksdk specific tests. OpenStack general tests
   should be submitted to the common 'hacking' module.
 - Pick numbers in the range O3xx. Find the current test with
   the highest allocated number and then pick the next value.
 - Keep the test method code in the source file ordered based
   on the O3xx value.
 - List the new rule in the top level HACKING.rst file
 - Add test cases for each new rule to openstack/tests/unit/test_hacking.py

"""

SETUPCLASS_RE = re.compile(r"def setUpClass\(")


@core.flake8ext
def assert_no_setupclass(logical_line):
    """Check for use of setUpClass

    O300
    """
    if SETUPCLASS_RE.match(logical_line):
        yield (0, "O300: setUpClass not allowed")


@core.flake8ext
def assert_no_deprecated_exceptions(logical_line, filename):
    """Check for use of deprecated cloud-layer exceptions

    0310
    """
    if filename.endswith(os.path.join('openstack', 'cloud', 'exc.py')):
        return

    for exception in (
        'OpenStackCloudCreateException',
        'OpenStackCloudTimeout',
        'OpenStackCloudHTTPError',
        'OpenStackCloudBadRequest',
        'OpenStackCloudURINotFound',
        'OpenStackCloudResourceNotFound',
    ):
        if re.search(rf'\b{exception}\b', logical_line):
            yield (0, 'O310: Use of deprecated Exception class')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class ClusterCertificate(resource.Resource):
    base_path = '/certificates'

    # capabilities
    allow_create = True
    allow_list = False
    allow_fetch = True

    #: The UUID of the bay.
    bay_uuid = resource.Body('bay_uuid')
    #: The UUID of the cluster.
    cluster_uuid = resource.Body('cluster_uuid', alternate_id=True)
    #: Certificate Signing Request (CSR) for authenticating client key.
    csr = resource.Body('csr')
    #: CA certificate for the bay/cluster.
    pem = resource.Body('pem')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.orchestration.v1 import stack_files as sf
from openstack import resource
from openstack.tests.unit import base

FAKE = {'stack_id': 'ID', 'stack_name': 'NAME'}


class TestStackFiles(base.TestCase):
    def test_basic(self):
        sot = sf.StackFiles()
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_list)

    def test_make_it(self):
        sot = sf.StackFiles(**FAKE)
        self.assertEqual(FAKE['stack_id'], sot.stack_id)
        self.assertEqual(FAKE['stack_name'], sot.stack_name)

    @mock.patch.object(resource.Resource, '_prepare_request')
    def test_get(self, mock_prepare_request):
        resp = mock.Mock()
        resp.json = mock.Mock(return_value={'file': 'file-content'})

        sess = mock.Mock()
        sess.get = mock.Mock(return_value=resp)

        sot = sf.StackFiles(**FAKE)

        req = mock.MagicMock()
        req.url = '/stacks/{stack_name}/{stack_id}/files'.format(
            stack_name=FAKE['stack_name'],
            stack_id=FAKE['stack_id'],
        )
        mock_prepare_request.return_value = req

        files = sot.fetch(sess)

        sess.get.assert_called_once_with(req.url)
        self.assertEqual({'file': 'file-content'}, files)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_rebuild_server
----------------------------------

Tests for the `rebuild_server` command.
"""

import uuid

from openstack import exceptions
from openstack.tests.unit import base
from openstack.tests.unit.cloud import fakes


class TestRebuildServer(base.TestCase):
    def setUp(self):
        super().setUp()
        self.server_id = str(uuid.uuid4())
        self.server_name = self.getUniqueString('name')
        self.fake_server = fakes.make_fake_server(
            self.server_id, self.server_name
        )
        self.rebuild_server = fakes.make_fake_server(
            self.server_id, self.server_name, 'REBUILD'
        )
        self.error_server = fakes.make_fake_server(
            self.server_id, self.server_name, 'ERROR'
        )

    def test_rebuild_server_rebuild_exception(self):
        """
        Test that an exception in the rebuild raises an exception in
        rebuild_server.
        """
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', self.server_id, 'action'],
                    ),
                    status_code=400,
                    validate=dict(
                        json={'rebuild': {'imageRef': 'a', 'adminPass': 'b'}}
                    ),
                ),
            ]
        )

        self.assertRaises(
            exceptions.SDKException,
            self.cloud.rebuild_server,
            self.fake_server['id'],
            "a",
            "b",
        )

        self.assert_calls()

    def test_rebuild_server_server_error(self):
        """
        Test that a server error while waiting for the server to rebuild
        raises an exception in rebuild_server.
        """
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', self.server_id, 'action'],
                    ),
                    json={'server': self.rebuild_server},
                    validate=dict(json={'rebuild': {'imageRef': 'a'}}),
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', self.server_id]
                    ),
                    json={'server': self.error_server},
                ),
            ]
        )
        self.assertRaises(
            exceptions.SDKException,
            self.cloud.rebuild_server,
            self.fake_server['id'],
            "a",
            wait=True,
        )

        self.assert_calls()

    def test_rebuild_server_timeout(self):
        """
        Test that a timeout while waiting for the server to rebuild raises an
        exception in rebuild_server.
        """
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', self.server_id, 'action'],
                    ),
                    json={'server': self.rebuild_server},
                    validate=dict(json={'rebuild': {'imageRef': 'a'}}),
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', self.server_id]
                    ),
                    json={'server': self.rebuild_server},
                ),
            ]
        )
        self.assertRaises(
            exceptions.ResourceTimeout,
            self.cloud.rebuild_server,
            self.fake_server['id'],
            "a",
            wait=True,
            timeout=0.001,
        )

        self.assert_calls(do_count=False)

    def test_rebuild_server_no_wait(self):
        """
        Test that rebuild_server with no wait and no exception in the
        rebuild call returns the server instance.
        """
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', self.server_id, 'action'],
                    ),
                    json={'server': self.rebuild_server},
                    validate=dict(json={'rebuild': {'imageRef': 'a'}}),
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'networks': []},
                ),
            ]
        )
        self.assertEqual(
            self.rebuild_server['status'],
            self.cloud.rebuild_server(self.fake_server['id'], "a")['status'],
        )

        self.assert_calls()

    def test_rebuild_server_with_admin_pass_no_wait(self):
        """
        Test that a server with an admin_pass passed returns the password
        """
        password = self.getUniqueString('password')
        rebuild_server = self.rebuild_server.copy()
        rebuild_server['adminPass'] = password

        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', self.server_id, 'action'],
                    ),
                    json={'server': rebuild_server},
                    validate=dict(
                        json={
                            'rebuild': {'imageRef': 'a', 'adminPass': password}
                        }
                    ),
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'networks': []},
                ),
            ]
        )
        self.assertEqual(
            password,
            self.cloud.rebuild_server(
                self.fake_server['id'], 'a', admin_pass=password
            )['adminPass'],
        )

        self.assert_calls()

    def test_rebuild_server_with_admin_pass_wait(self):
        """
        Test that a server with an admin_pass passed returns the password
        """
        password = self.getUniqueString('password')
        rebuild_server = self.rebuild_server.copy()
        rebuild_server['adminPass'] = password

        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', self.server_id, 'action'],
                    ),
                    json={'server': rebuild_server},
                    validate=dict(
                        json={
                            'rebuild': {'imageRef': 'a', 'adminPass': password}
                        }
                    ),
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', self.server_id]
                    ),
                    json={'server': self.rebuild_server},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', self.server_id]
                    ),
                    json={'server': self.fake_server},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'networks': []},
                ),
            ]
        )

        self.assertEqual(
            password,
            self.cloud.rebuild_server(
                self.fake_server['id'], 'a', admin_pass=password, wait=True
            )['adminPass'],
        )

        self.assert_calls()

    def test_rebuild_server_wait(self):
        """
        Test that rebuild_server with a wait returns the server instance when
        its status changes to "ACTIVE".
        """
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', self.server_id, 'action'],
                    ),
                    json={'server': self.rebuild_server},
                    validate=dict(json={'rebuild': {'imageRef': 'a'}}),
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', self.server_id]
                    ),
                    json={'server': self.rebuild_server},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', self.server_id]
                    ),
                    json={'server': self.fake_server},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'networks': []},
                ),
            ]
        )
        self.assertEqual(
            'ACTIVE',
            self.cloud.rebuild_server(self.fake_server['id'], 'a', wait=True)[
                'status'
            ],
        )

        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.baremetal.v1 import _common
from openstack import resource


class InspectionRule(resource.Resource):
    resources_key = 'inspection_rules'
    base_path = '/inspection_rules'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    allow_patch = True
    commit_method = 'PATCH'

    _query_mapping = resource.QueryParameters(
        'detail',
        fields={'type': _common.fields_type},
    )

    # Inspection rules is available since 1.96
    _max_microversion = '1.96'
    #: The actions to be executed when the rule conditions are met.
    actions = resource.Body('actions', type=list)
    #: A brief explanation about the inspection rule.
    description = resource.Body('description')
    #: The conditions under which the rule should be triggered.
    conditions = resource.Body('conditions', type=list)
    #: Timestamp at which the resource was created.
    created_at = resource.Body('created_at')
    #: A list of relative links. Includes the self and bookmark links.
    links = resource.Body('links', type=list)
    #: Specifies the phase when the rule should run, defaults to 'main'.
    phase = resource.Body('phase')
    #: Specifies the rule's precedence level during execution.
    priority = resource.Body('priority')
    #: Indicates whether the rule contains sensitive information.
    sensitive = resource.Body('sensitive', type=bool)
    #: Timestamp at which the resource was last updated.
    updated_at = resource.Body('updated_at')
    #: The UUID of the resource.
    id = resource.Body('uuid', alternate_id=True)
